# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import onnxscript.ir as ir
from onnxscript.rewriter import _fusion_utils, _ir_utils, pattern

"""
RMS Normalization: ONNX Opset 23 op
See: https://onnx.ai/onnx/operators/onnx__RMSNormalization.html#l-onnx-doc-rmsnormalization


Key points for the fusion optimization:
* Input and scale are allowed to be of different types.
* The normalization of the input can be done in a different precision than the input type,
indicated by stash_type.
* Input (x) must be: float or double or float16 or bfloat16
* Scale must be: float or double or float16 or bfloat16
"""

float_types = frozenset(
    [
        ir.DataType.FLOAT,
        ir.DataType.FLOAT16,
        ir.DataType.BFLOAT16,
        ir.DataType.DOUBLE,
    ]
)
fp_float_types = frozenset([ir.DataType.FLOAT, ir.DataType.DOUBLE])


class RmsNormFusion(pattern.RewriteRuleClassBase):
    def __init__(self, name: str, mul_order: bool):
        super().__init__(name)
        self._mul_order = mul_order

    def pattern(self, op, x, scale, epsilon, compute_dtype, target_dtype):
        x = pattern.OrValue([op.Cast(x, to=compute_dtype), x])
        x_square = op.Pow(x, 2.0)
        mean_square = op.ReduceMean(x_square, [-1], keepdims=1, noop_with_empty_axes=0)
        mean_square_plus_epsilon = op.Add(mean_square, epsilon)
        rms = op.Sqrt(mean_square_plus_epsilon)
        reciprocal_rms = op.Reciprocal(rms)
        normalized = op.Mul(x, reciprocal_rms)
        normalized = pattern.OrValue([op.Cast(normalized, to=target_dtype), normalized])
        # Workaround: limitation in pattern matcher doesn't support OrValue for return value (last node in pattern)
        if self._mul_order:
            return op.Mul(normalized, scale)
        else:
            return op.Mul(scale, normalized)

    def check(
        self, op, x, scale, epsilon, compute_dtype, target_dtype, **_
    ) -> pattern.MatchResult:  # type: ignore[name-defined]
        """Check if the pattern matches conditions for use of SimplifiedLayerNormalization op."""
        check_result = pattern.MatchResult()
        # epsilon must be a scalar
        epsilon_value = _ir_utils.get_singleton_value(epsilon)
        if not isinstance(epsilon_value, float):  # TODO: support other types
            return check_result.fail("Epsilon is not a float value.", epsilon)
        if x.dtype not in float_types:
            return check_result.fail("Input is not a supported float type.", x)
        if scale.dtype not in float_types:
            return check_result.fail("Scale is not a supported float type.", scale)
        self._stash_dtype = compute_dtype.as_int() if compute_dtype is not None else x.dtype
        if self._stash_dtype not in fp_float_types:
            # TODO: ONNX documentation does not specify restrictions on stash_type, though
            # ORT's SimplifiedLayerNormalization requires it to be float or double.
            return check_result.fail("Normalization precision is not a float or double type.")
        # target_dtype is guaranteed to be the same as scale type in a well-typed input
        # for Mul(scale, normalized) to work. There is no need to check it here for a well-typed input.
        # TODO (rama): Consider adding checks to protect against incorrectly typed models:
        return check_result

    def rewrite(self, op, x, scale, epsilon, **_):
        # Note: ORT's SimplifiedLayerNormalization was placed in onnx domain by mistake.
        # No need to use com.microsoft domain here; but this is a custom op in ORT.
        return op.RMSNormalization(
            x,
            scale,
            axis=-1,
            epsilon=_ir_utils.get_singleton_value(epsilon),
            stash_type=self._stash_dtype,
        )


_rule1 = RmsNormFusion.rule("RmsNormFusion1", mul_order=True)
_rule2 = RmsNormFusion.rule("RmsNormFusion2", mul_order=False)

rms_normalization_rules = [_rule1, _rule2]

rms_normalization_ruleset = pattern.RewriteRuleSet(rms_normalization_rules)

fuse_rms_normalization = _fusion_utils.apply_fusion_rules(rms_normalization_ruleset)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Utilities for interfacing with onnx C APIs."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Callable, TypeVar

from onnxscript import ir

if TYPE_CHECKING:
    import onnx


logger = logging.getLogger(__name__)
# Temporarily remove initializers larger than this size to keep model size down
# for the onnx.shape_inference call because it needs to serialize the model
_BIG_TENSOR_SIZE_LIMIT = 1000  # 1KB
_R = TypeVar("_R")


def call_onnx_api(func: Callable[[onnx.ModelProto], _R], model: ir.Model) -> _R:
    """Call an ONNX C API function by temporarily removing initializers.

    This is necessary because the ONNX C API does not support large models
    with initializers that have large tensor values. The input model is left
    unchanged no matter the call succeeds or not.

    Args:
        func: Partially applied function that takes a model proto and returns anything.
        model: The IR model to pass to the API function.

    Returns:
        The resulting ModelProto that contains the result of the API call.
    """

    # Store the original initializer values so they can be restored
    initializer_values = tuple(model.graph.initializers.values())
    tensors = {v.name: v.const_value for v in initializer_values}
    original_inputs_len = len(model.graph.inputs)

    # Turn the initializers into inputs and clear the initializers
    # to limit the model size
    for initializer in initializer_values:
        # Make sure the initializer has its shape/type set
        assert initializer.const_value is not None
        if initializer.shape is None:
            initializer.shape = initializer.const_value.shape  # type: ignore[assignment]
        if initializer.dtype is None:
            initializer.dtype = initializer.const_value.dtype
        if initializer not in model.graph.inputs:
            model.graph.inputs.append(initializer)
        if initializer.const_value.size > _BIG_TENSOR_SIZE_LIMIT:
            # Temporarily remove the initializer value to reduce model size
            # for onnx.shape_inference
            initializer.const_value = None
            assert initializer.name is not None
            model.graph.initializers.pop(initializer.name)

    proto = ir.serde.serialize_model(model)

    try:
        # Call the ONNX C API function
        result = func(proto)
    finally:
        # Restore the original initializer values so the model is unchanged
        for initializer in initializer_values:
            initializer.const_value = tensors[initializer.name]
            model.graph.register_initializer(initializer)

        # Restore the original inputs
        inputs = model.graph.inputs[:original_inputs_len]
        model.graph.inputs.clear()
        model.graph.inputs.extend(inputs)

    return result



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import Mapping, Sequence

import onnx_ir as ir
from onnx_ir._cloner import Cloner


def instantiate(
    graph: ir.Graph,
    inputs: Sequence[ir.Value | None],
    attributes: Mapping[str, ir.Attr],
    *,
    prefix: str = "",
) -> tuple[list[ir.Node], list[ir.Value | None]]:
    """Instantiate (inline) a graph, substituting inputs and attributes.

    Args:
        graph: The graph to instantiate.
        inputs: Actual input values to bind to the graph's formal parameters.
        attributes: Attribute values to substitute for reference attributes.
        prefix: Optional prefix to prepend to node and output names.

    Returns:
        A tuple of (nodes, outputs) where nodes are the cloned graph body
        and outputs are the values corresponding to the graph's outputs.
    """
    formal_inputs = graph.inputs
    if len(inputs) > len(formal_inputs):
        raise ValueError(
            f"Too many inputs: got {len(inputs)}, "
            f"but graph has {len(formal_inputs)} parameters."
        )
    value_map: dict[ir.Value, ir.Value | None] = dict(zip(formal_inputs, inputs))

    def rename(node: ir.Node) -> None:
        if prefix:
            if node.name:
                node.name = prefix + node.name
            for output in node.outputs:
                if output is not None and output.name:
                    output.name = prefix + output.name

    cloner = Cloner(
        attr_map=attributes,
        value_map=value_map,
        metadata_props={},
        post_process=rename,
        resolve_ref_attrs=True,
        allow_outer_scope_values=True,
    )
    nodes = [cloner.clone_node(n) for n in graph]
    outputs = [value_map.get(v) for v in graph.outputs]
    return nodes, outputs



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Absorbs affine operation into convolution (best effort):
- Conv(Mul(Add(x))) -> Conv (only conv without padding can be fused)
- Add(Mul(Conv)) -> Conv (for all convolutions)
"""

from __future__ import annotations

import numpy as np
import onnx_ir as ir

from onnxscript.rewriter import pattern
from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._ir_utils import get_const_value, get_singleton_value


class _ConvAffineFusionBase(pattern.RewriteRuleClassBase):
    def check(
        self,
        context,
        x: ir.Value,
        w: ir.Value,
        b: ir.Value,
        scale: ir.Value,
        offset: ir.Value,
        conv_out: ir.Value,
    ) -> MatchResult:
        check_result = MatchResult()
        if get_const_value(w) is None:
            return check_result.fail("The weight of Conv should be constant")
        if get_const_value(b) is None:
            return check_result.fail("The bias of Conv should be constant")
        if get_singleton_value(scale) is None:
            return check_result.fail("Operand for Mul should be constant scalar value")
        if get_singleton_value(offset) is None:
            return check_result.fail("Operand for Add should be constant scalar value")
        return check_result


class AffineConvFusion(_ConvAffineFusionBase):
    """Pattern: scalar Mul + scalar Add + Conv (1x1) --> Conv(1x1)"""

    def pattern(
        self, op, x: ir.Value, w: ir.Value, b: ir.Value, scale: ir.Value, offset: ir.Value
    ) -> ir.Value:
        return op.Conv(
            x * scale + offset,
            w,
            b,
            pads=[0, 0, 0, 0],
            _allow_other_attributes=True,
            _outputs=["conv_out"],
        )

    def rewrite(
        self,
        op: ir.tape.Tape,
        x: ir.Value,
        w: ir.Value,
        b: ir.Value,
        scale: ir.Value,
        offset: ir.Value,
        conv_out: ir.Value,
    ) -> ir.Value:
        scale_value = scale.const_value.numpy()
        offset_value = offset.const_value.numpy()
        w_value = w.const_value.numpy()
        b_value = b.const_value.numpy()
        scaled_w_value = op.initializer(ir.tensor(w_value * scale_value), w.name + "_scaled")
        offset_bias = ir.tensor(
            b_value + np.sum(w_value * offset_value, axis=(1, 2, 3), keepdims=False)
        )
        offset_bias = op.initializer(offset_bias, b.name + "_offset")
        conv_attributes = conv_out.producer().attributes
        return op.Conv(x, scaled_w_value, offset_bias, **conv_attributes)


class ConvAffineFusion(_ConvAffineFusionBase):
    """Pattern: Conv + scalar Mul + scalar Add --> Conv(1x1)"""

    def pattern(
        self, op, x: ir.Value, w: ir.Value, b: ir.Value, scale: ir.Value, offset: ir.Value
    ) -> ir.Value:
        return (
            op.Conv(x, w, b, _allow_other_attributes=True, _outputs=["conv_out"]) * scale
            + offset
        )

    def rewrite(
        self,
        op: ir.tape.Tape,
        x: ir.Value,
        w: ir.Value,
        b: ir.Value,
        scale: ir.Value,
        offset: ir.Value,
        conv_out: ir.Value,
    ) -> ir.Value:
        scale_value = scale.const_value.numpy()
        offset_value = offset.const_value.numpy()
        w_value = w.const_value.numpy()
        b_value = b.const_value.numpy()
        scaled_w_weight = op.initializer(ir.tensor(w_value * scale_value), w.name + "_scaled")
        offset_bias = ir.tensor(b_value * scale_value + offset_value)
        offset_bias = op.initializer(offset_bias, b.name + "_offset")
        conv_attributes = conv_out.producer().attributes
        return op.Conv(x, scaled_w_weight, offset_bias, **conv_attributes)


affine_conv_fusion_rule = AffineConvFusion().rule()
conv_affine_fusion_rule = ConvAffineFusion().rule()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

import onnx.parser
import onnx.shape_inference
import parameterized

from onnxscript import ir
from onnxscript.rewriter.rules.common import _broadcast_to_matmul


def _infer_shapes(model: ir.Model) -> ir.Model:
    """Run shape inference on the IR model."""
    # TODO: Update when shape inference is supported on the IR
    return ir.serde.deserialize_model(
        onnx.shape_inference.infer_shapes(ir.serde.serialize_model(model))
    )


class TwoReshapesMatMulReshapeTest(unittest.TestCase):
    def test_reshape_matmul_reshape_replace_when_nd_inputs_are_broadcastable(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[1, 4, 512, 512] input_x, float[1, 4, 512, 64] input_y) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    @parameterized.parameterized.expand(
        [
            (
                "0d",
                [],
                [1, 1],
                [],
                [1, 1],
                [1, 1],
                [1, 1],
            ),
            (
                "x_1d",
                [4],
                [1, 4],
                [4, 2],
                [4, 2],
                [1, 2],
                [1, 2],
            ),
            (
                "y_1d",
                [1, 4],
                [1, 4],
                [2],
                [4, 2],
                [1, 2],
                [1, 2],
            ),
            (
                "both_1d",
                [2],
                [1, 2],
                [2],
                [2, 1],
                [],
                [],
            ),
        ]
    )
    def test_reshape_matmul_reshape_does_not_replace_when_output_sizes_do_not_match(
        self,
        _: str,
        input_x_shape: list[int],
        shape_a: list[int],
        input_y_shape: list[int],
        shape_b: list[int],
        output_shape: list[int],
        shape_c: list[int],
    ):
        model_proto = onnx.parser.parse_model(
            f"""
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float{input_x_shape} input_x, float{input_y_shape} input_y) => (float{output_shape} output)
            {{
                shape_a = Constant<value: tensor = int64[{len(shape_a)}] {{ {", ".join(str(i) for i in shape_a)} }}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[{len(shape_b)}] {{ {", ".join(str(i) for i in shape_b)} }}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[{len(shape_c)}] {{ {", ".join(str(i) for i in shape_c)} }}>()
                output = Reshape (matmul, shape_c)
            }}
            """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 7)
        model = _infer_shapes(model)
        self.assertEqual(model.graph.outputs[0].shape, output_shape)

    def test_reshape_matmul_reshape_replace_when_nd_inputs_are_broadcastable_in_nested_function(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17, "pkg.custom": 1]>
            agraph (float[1, 4, 512, 512] input_x, float[1, 4, 512, 64] input_y) => (float[1, 4, 512, 64] output)
            {
                output = pkg.custom.afunction (input_x, input_y)
            }
            <domain: "pkg.custom", opset_import: [ "" : 17]>
            afunction (input_x, input_y) => (output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        # Hack to put value_info in since parser does not support this experimental naming format
        model_proto.graph.value_info.append(
            onnx.helper.make_tensor_value_info(
                "pkg.custom::afunction/input_x",
                onnx.TensorProto.FLOAT,
                [1, 4, 512, 512],
            )
        )
        model_proto.graph.value_info.append(
            onnx.helper.make_tensor_value_info(
                "pkg.custom::afunction/input_y", onnx.TensorProto.FLOAT, [1, 4, 512, 64]
            )
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.functions), 1)
        self.assertEqual(len(model.functions[("pkg.custom", "afunction", "")]), 4)
        self.assertEqual(
            model.functions[("pkg.custom", "afunction", "")][-1].op_type, "MatMul"
        )

    def test_reshape_matmul_reshape_remain_when_input_last_dim_and_second_last_dim_not_matched(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[512, 512, 4] input_x, float[4, 64, 512] input_y) => (float[1, 4, 512, 64] output)
            {
                shape_a = Constant<value: tensor = int64[3] {4, 512, 512}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[3] {4, 512, 64}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {1, 4, 512, 64}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 7)

    def test_reshape_matmul_reshape_remain_one_reshape_when_inputs_are_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 8, 512, 64] input_x, float[4, 4, 64, 512] input_y) => (float[2, 8, 512, 512] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 8, 512, 64}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[4] {2, 8, 64, 512}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {2, 8, 512, 512}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model_proto = onnx.shape_inference.infer_shapes(model_proto)
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        # subset pattern matched
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_matmul_reshape_replace_when_inputs_are_broadcastable_with_one_in_dims(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 8, 512, 64] input_x, float[1, 1, 2, 8, 64, 512] input_y) => (float[1, 1, 2, 8, 512, 512] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 8, 512, 64}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[6] {1, 1, 2, 8, 64, 512}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[6] {1, 1, 2, 8, 512, 512}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_matmul_reshape_replace_when_first_input_is_one_dimension_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[4] input_x, float[2, 3, 4, 5] input_y) => (float[2, 3, 5] output)
            {
                shape_a = Constant<value: tensor = int64[2] {1, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[4] {2, 3, 4, 5}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[3] {2, 3, 5}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_matmul_reshape_replace_when_first_input_is_one_dimension_and_second_isexpanded_alike_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[5] input_x, float[5, 1] input_y) => (float[1] output)
            {
                shape_a = Constant<value: tensor = int64[2] {1, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[2] {5, 1}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[1] {1}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_matmul_reshape_remain_when_first_input_is_one_dimension_and_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[8] input_x, float[2, 3, 4, 5] input_y) => (float[2, 3, 2, 5] output)
            {
                shape_a = Constant<value: tensor = int64[2] {2, 4}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[4] {2, 3, 4, 5}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {2, 3, 2, 5}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 7)

    def test_reshape_matmul_reshape_replace_when_second_input_is_one_dimension_and_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 4, 5] input_x, float[5] input_y) => (float[2, 3, 4] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 4, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[2] {5, 1}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[3] {2, 3, 4}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 4)

    def test_reshape_matmul_reshape_remain_one_reshape_when_second_input_is_one_dimension_and_not_broadcastable(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 4, 5] input_x, float[10] input_y) => (float[2, 3, 4, 2] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 4, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[2] {5, 2}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {2, 3, 4, 2}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model_proto = onnx.shape_inference.infer_shapes(model_proto)
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        # subset pattern matched
        self.assertEqual(count, 1)
        self.assertEqual(len(model.graph), 5)

    def test_reshape_matmul_reshape_remain_when_output_is_not_matmul_broadcasted(
        self,
    ):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[2, 3, 4, 5] input_x, float[5, 8] input_y) => (float[2, 4, 6, 4] output)
            {
                shape_a = Constant<value: tensor = int64[4] {2, 3, 4, 5}>()
                reshape_x = Reshape (input_x, shape_a)
                shape_b = Constant<value: tensor = int64[2] {5, 2}>()
                reshape_y = Reshape (input_y, shape_b)
                matmul = MatMul (reshape_x, reshape_y)
                shape_c = Constant<value: tensor = int64[4] {2, 4, 6, 4}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 0)
        self.assertEqual(len(model.graph), 7)


class OneReshapeMatMulReshapeTest(unittest.TestCase):
    def test_reshape_matmul_reshape_replace_when_nd_inputs_are_broadcastable(self):
        model_proto = onnx.parser.parse_model(
            """
            <ir_version: 7, opset_import: [ "" : 17]>
            agraph (float[1, 512, 4096] input_x, float[4096, 4096] input_y) => (float[1, 512, 4096] output)
            {
                shape_a = Constant<value: tensor = int64[3] {1, 512, 4096}>()
                reshape_x = Reshape (input_x, shape_a)
                matmul = MatMul (reshape_x, input_y)
                shape_c = Constant<value: tensor = int64[3] {1, 512, 4096}>()
                output = Reshape (matmul, shape_c)
            }
        """
        )
        model = ir.serde.deserialize_model(model_proto)
        count = _broadcast_to_matmul.rules.apply_to_model(model)
        self.assertEqual(count, 1)
        # The constant nodes are not removed. They should be removed by a subsequent DCE in optimizer.
        self.assertEqual(len(model.graph), 3)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# ruff: noqa: TID251

import os
import textwrap
from typing import Iterator

import numpy as np
import onnx
import onnx.numpy_helper
from onnx.backend.test import __file__ as backend_folder

from onnxscript.backend import onnx_export


def assert_almost_equal_string(expected, value):
    """Compares two arrays knowing they contain strings.
    Raises an exception if the test fails.

    Args:
        expected: expected array
        value: value
    """

    def is_float(x):  # pylint: disable=unused-argument
        try:
            return True
        except ValueError:  # pragma: no cover
            return False

    if all(map(is_float, expected.ravel())):
        expected_float = expected.astype(np.float32)
        value_float = value.astype(np.float32)
        np.testing.assert_almost_equal(expected_float, value_float)
    else:
        np.testing.assert_almost_equal(expected, value)


class OnnxBackendTest:
    """Definition of a backend test. It starts with a folder,
    in this folder, one onnx file must be there, then a subfolder
    for each test to run with this model.

    Args:
        folder: test folder
        onnx_path: onnx file
        onnx_model: loaded onnx file
        tests: list of test
    """

    @staticmethod
    def _sort(filenames):
        temp = []
        for f in filenames:
            name = os.path.splitext(f)[0]
            i = name.split("_")[-1]
            temp.append((int(i), f))
        temp.sort()
        return [_[1] for _ in temp]

    @staticmethod
    def _read_proto_from_file(full):
        if not os.path.exists(full):
            raise FileNotFoundError(f"File not found: {full!r}.")  # pragma: no cover
        with open(full, "rb") as f:
            serialized = f.read()
        try:
            loaded = onnx.numpy_helper.to_array(onnx.load_tensor_from_string(serialized))
        except Exception as e:  # pylint: disable=W0703
            seq = onnx.SequenceProto()
            try:
                seq.ParseFromString(serialized)
                loaded = onnx.numpy_helper.to_list(seq)  # type: ignore[assignment]
            except Exception:  # pylint: disable=W0703
                try:
                    loaded = onnx.load_model_from_string(serialized)  # type: ignore[assignment]
                except Exception:
                    raise RuntimeError(
                        f"Unable to read {full!r}, error is {e}, "
                        f"content is {serialized[:100]!r}."
                    ) from e
        return loaded

    @staticmethod
    def _load(folder, names):
        res = []
        for name in names:
            full = os.path.join(folder, name)
            new_tensor = OnnxBackendTest._read_proto_from_file(full)
            if isinstance(new_tensor, (np.ndarray, onnx.ModelProto, list)):
                t = new_tensor
            elif isinstance(new_tensor, onnx.TensorProto):
                t = onnx.numpy_helper.to_array(new_tensor)
            else:
                raise RuntimeError(  # noqa: TRY004
                    f"Unexpected type {type(new_tensor)!r} for {full!r}."
                )
            res.append(t)
        return res

    def __repr__(self):
        """Usual"""
        return f"{self.__class__.__name__}({self.folder!r})"

    def __init__(self, folder):
        if not os.path.exists(folder):
            raise FileNotFoundError(f"Unable to find folder {folder!r}.")  # pragma: no cover
        content = os.listdir(folder)
        onx = [c for c in content if os.path.splitext(c)[-1] in {".onnx"}]
        if len(onx) != 1:
            raise ValueError(  # pragma: no cover
                f"There is more than one onnx file in {folder!r} ({onx!r})."
            )
        self.folder = folder
        self.onnx_path = os.path.join(folder, onx[0])
        self.onnx_model = onnx.load(self.onnx_path)

        self.tests = []
        for sub in content:
            full = os.path.join(folder, sub)
            if os.path.isdir(full):
                pb = [c for c in os.listdir(full) if os.path.splitext(c)[-1] in {".pb"}]
                inputs = OnnxBackendTest._sort(c for c in pb if c.startswith("input_"))
                outputs = OnnxBackendTest._sort(c for c in pb if c.startswith("output_"))

                t = dict(
                    inputs=OnnxBackendTest._load(full, inputs),
                    outputs=OnnxBackendTest._load(full, outputs),
                )
                self.tests.append(t)

    @property
    def name(self):
        """Returns the test name."""
        return os.path.split(self.folder)[-1]

    def __len__(self):
        """Returns the number of tests."""
        return len(self.tests)

    def _compare_results(self, index, i, e, o, decimal=None):
        """Compares the expected output and the output produced
        by the runtime. Raises an exception if not equal.

        Args:
            index: test index
            i: output index
            e: expected output
            o: output
            decimal: precision
        """
        if isinstance(e, np.ndarray):
            if isinstance(o, np.ndarray):
                if decimal is None:
                    if e.dtype == np.float32:
                        deci = 6
                    elif e.dtype == np.float64:
                        deci = 12
                    else:
                        deci = 7
                else:
                    deci = decimal
                if e.dtype == np.object_:
                    try:
                        assert_almost_equal_string(e, o)
                    except AssertionError as ex:
                        raise AssertionError(  # pragma: no cover
                            f"Output {i} of test {index} in folder {self.folder} failed."
                        ) from ex
                else:
                    try:
                        np.testing.assert_almost_equal(e, o, decimal=deci)
                    except AssertionError as ex:
                        raise AssertionError(
                            f"Output {i} of test {index} in folder {self.folder} failed."
                        ) from ex
            elif hasattr(o, "is_compatible"):
                # A shape
                if e.dtype != o.dtype:
                    raise AssertionError(
                        f"Output {i} of test {index} in folder "
                        f"{self.folder} failed (e.dtype={e.dtype}, o={o})."
                    )
                if not o.is_compatible(e.shape):
                    raise AssertionError(  # pragma: no cover
                        f"Output {i} of test {index} in folder "
                        f"{self.folder} failed (e.shape={e.shape}, o={o})."
                    )
        else:
            raise NotImplementedError(f"Comparison not implemented for type {type(e)!r}.")

    def is_random(self):
        """Returns whether the test is random."""
        return "bernoulli" in self.folder

    def run(self, load_fct, run_fct, index=None, decimal=None):
        """Executes a tests or all tests if index is None.
        The function crashes if the tests fails.

        Args:
            load_fct: loading function, takes a loaded onnx graph, and
                returns an object
            run_fct: running function, takes the result of previous
                function, the inputs, and returns the outputs
            index: index of the test to run or all.
            decimal: requested precision to compare results
        """
        if index is None:
            for i in range(len(self)):
                self.run(load_fct, run_fct, index=i, decimal=decimal)
            return

        obj = load_fct(self.onnx_model)

        got = run_fct(obj, *self.tests[index]["inputs"])
        expected = self.tests[index]["outputs"]
        if len(got) != len(expected):
            raise AssertionError(  # pragma: no cover
                f"Unexpected number of output (test {index}, folder {self.folder}), "
                f"got {len(got)}, expected {len(expected)}."
            )
        for i, (e, o) in enumerate(zip(expected, got)):
            if self.is_random():
                if e.dtype != o.dtype:
                    raise AssertionError(
                        f"Output {i} of test {index} in folder "
                        f"{self.folder} failed (type mismatch {e.dtype} != {o.dtype})."
                    )
                if e.shape != o.shape:
                    raise AssertionError(
                        f"Output {i} of test {index} in folder "
                        f"{self.folder} failed (shape mismatch {e.shape} != {o.shape})."
                    )
            else:
                self._compare_results(index, i, e, o, decimal=decimal)

    def to_python(self):
        """Returns a python code equivalent to the ONNX test.

        Returns:
            code
        """
        rows = []
        code = onnx_export.export2onnx(self.onnx_model)  # type: ignore[attr-defined]
        lines = code.split("\n")
        lines = [
            line
            for line in lines
            if not line.strip().startswith("print") and not line.strip().startswith("# ")
        ]
        rows.append(textwrap.dedent("\n".join(lines)))
        rows.append("oinf = OnnxInference(onnx_model)")
        for test in self.tests:
            rows.append("xs = [")
            for inp in test["inputs"]:
                rows.append(textwrap.indent(f"{inp!r},", "    " * 2))
            rows.append("]")
            rows.append("ys = [")
            for out in test["outputs"]:
                rows.append(textwrap.indent(f"{out!r},", "    " * 2))
            rows.append("]")
            rows.append("feeds = {n: x for n, x in zip(oinf.input_names, xs)}")
            rows.append("got = oinf.run(feeds)")
            rows.append("goty = [got[k] for k in oinf.output_names]")
            rows.append("for y, gy in zip(ys, goty):")
            rows.append("    self.assertEqualArray(y, gy)")
            rows.append("")
        code = "\n".join(rows)
        final = "\n".join([f"def {self.name}(self):", textwrap.indent(code, "    ")])
        return final


def enumerate_onnx_tests(series, fct_filter=None) -> Iterator[OnnxBackendTest]:
    """Collects test from a sub folder of `onnx/backend/test`.
    Works as an enumerator to start processing them
    without waiting or storing too much of them.

    Args:
        series: which subfolder to load, possible values: (`'node'`,
            ...)
        fct_filter: function `lambda testname: boolean` to load or skip
            the test, None for all

    Yields:
        list of @see cl OnnxBackendTest
    """
    root = os.path.dirname(backend_folder)
    sub = os.path.join(root, "data", series)
    if not os.path.exists(sub):
        raise FileNotFoundError(
            f"Unable to find series of tests in {root!r}, subfolders:\n"
            + "\n".join(os.listdir(root))
        )
    tests = os.listdir(sub)
    for t in tests:
        if fct_filter is not None and not fct_filter(t):
            continue
        folder = os.path.join(sub, t)
        content = os.listdir(folder)
        onx = [c for c in content if os.path.splitext(c)[-1] in {".onnx"}]
        if len(onx) == 1:
            yield OnnxBackendTest(folder)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

"""
A one-layer SmolLM model test case, with inputs: input_ids, position_ids, and pask key/values.
This is an onnxscript version of the model.
"""

import numpy
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18
from onnxscript.onnx_types import FLOAT, INT64


def make_model(
    model_layers_0_input_layernorm_weight,
    model_layers_0_post_attention_layernorm_weight,
    model_norm_weight,
    lm_head_weight,
    model_layers_0_self_attn_q_proj_weight,
    model_layers_0_self_attn_k_proj_weight,
    model_layers_0_self_attn_v_proj_weight,
    model_layers_0_self_attn_o_proj_weight,
    model_layers_0_mlp_gate_proj_weight,
    model_layers_0_mlp_up_proj_weight,
    model_layers_0_mlp_down_proj_weight,
    model_rotary_emb_inv_freq,
):
    @script()
    def main_graph(
        input_ids: INT64[1, 30],
        position_ids: INT64[1, 30],
        past_key_values_0_0: FLOAT[1, 32, 16, 64],
        past_key_values_0_1: FLOAT[1, 32, 16, 64],
    ) -> (FLOAT[1, 30, 49152], FLOAT[1, 32, 46, 64], FLOAT[1, 32, 46, 64]):
        embedding = opset18.Gather(lm_head_weight, input_ids, axis=0)
        val_2 = opset18.CastLike(1.0, 46)
        arange = opset18.Range(16, 46, val_2)
        val_5 = opset18.Cast(-3.4028235e38, to=1)
        val_7 = opset18.Cast([30, 47], to=7)
        full = opset18.Expand(val_5, val_7)
        diagonal__1 = opset18.Constant(value_int=1)
        triu = opset18.Trilu(full, diagonal__1, upper=1)
        val_10 = opset18.CastLike(0.0, 47)
        val_11 = opset18.CastLike(1.0, 47)
        arange_1 = opset18.Range(val_10, 47, val_11)
        val_13 = opset18.Cast([-1, 1], to=7)
        view = opset18.Reshape(arange, val_13, allowzero=0)
        gt = arange_1 > view
        convert_element_type_default = opset18.Cast(gt, to=1)
        mul = triu * convert_element_type_default
        dim__2 = opset18.Constant(value_ints=[0])
        dim_0__2 = opset18.Cast(dim__2, to=7)
        unsqueeze = opset18.Unsqueeze(model_rotary_emb_inv_freq, dim_0__2)
        val_15 = opset18.Cast(0, to=7)
        val_16 = opset18.Constant(value_ints=[-1])
        val_17 = opset18.Reshape(val_15, val_16, allowzero=0)
        val_19 = opset18.Cast(9223372036854775807, to=7)
        val_20 = opset18.Constant(value_ints=[-1])
        val_21 = opset18.Reshape(val_19, val_20, allowzero=0)
        val_23 = opset18.Cast(1, to=7)
        val_24 = opset18.Constant(value_ints=[-1])
        val_25 = opset18.Reshape(val_23, val_24, allowzero=0)
        val_26 = opset18.Constant(value_ints=[1])
        slice_1 = opset18.Slice(unsqueeze, val_17, val_21, val_25, val_26)
        dim__3 = opset18.Constant(value_ints=[2])
        dim_0__3 = opset18.Cast(dim__3, to=7)
        unsqueeze_1 = opset18.Unsqueeze(slice_1, dim_0__3)
        _to_copy = opset18.Cast(unsqueeze_1, to=1)
        size_0__4 = opset18.Cast([1, -1, 1], to=7)
        size_1__4 = opset18.Abs(size_0__4)
        expand = opset18.Expand(_to_copy, size_1__4)
        val_28 = opset18.Cast(0, to=7)
        val_29 = opset18.Constant(value_ints=[-1])
        val_30 = opset18.Reshape(val_28, val_29, allowzero=0)
        val_31 = opset18.Cast(9223372036854775807, to=7)
        val_32 = opset18.Constant(value_ints=[-1])
        val_33 = opset18.Reshape(val_31, val_32, allowzero=0)
        val_34 = opset18.Cast(0, to=7)
        val_35 = opset18.Constant(value_ints=[-1])
        val_36 = opset18.Reshape(val_34, val_35, allowzero=0)
        val_37 = opset18.Constant(value_ints=[1])
        slice_2 = opset18.Slice(position_ids, val_30, val_33, val_36, val_37)
        dim__5 = opset18.Constant(value_ints=[1])
        dim_0__5 = opset18.Cast(dim__5, to=7)
        unsqueeze_2 = opset18.Unsqueeze(slice_2, dim_0__5)
        val_38 = opset18.Cast(0, to=7)
        val_39 = opset18.Constant(value_ints=[-1])
        val_40 = opset18.Reshape(val_38, val_39, allowzero=0)
        val_41 = opset18.Cast(9223372036854775807, to=7)
        val_42 = opset18.Constant(value_ints=[-1])
        val_43 = opset18.Reshape(val_41, val_42, allowzero=0)
        val_45 = opset18.Cast(2, to=7)
        val_46 = opset18.Constant(value_ints=[-1])
        val_47 = opset18.Reshape(val_45, val_46, allowzero=0)
        val_48 = opset18.Constant(value_ints=[1])
        slice_3 = opset18.Slice(unsqueeze_2, val_40, val_43, val_47, val_48)
        _to_copy_1 = opset18.Cast(slice_3, to=1)
        _to_copy_2 = opset18.Cast(expand, to=1)
        _to_copy_3 = opset18.Cast(_to_copy_1, to=1)
        size_0__6 = opset18.Cast([1, 32, 1], to=7)
        size_1__6 = opset18.Abs(size_0__6)
        expand_1 = opset18.Expand(_to_copy_2, size_1__6)
        val_50 = opset18.Cast([1, 32, 1], to=7)
        view_1 = opset18.Reshape(expand_1, val_50, allowzero=0)
        size_0__7 = opset18.Cast([1, 1, 30], to=7)
        size_1__7 = opset18.Abs(size_0__7)
        expand_2 = opset18.Expand(_to_copy_3, size_1__7)
        val_52 = opset18.Cast([1, 1, 30], to=7)
        view_2 = opset18.Reshape(expand_2, val_52, allowzero=0)
        bmm = view_1 @ view_2
        val_54 = opset18.Cast([1, 32, 30], to=7)
        view_3 = opset18.Reshape(bmm, val_54, allowzero=0)
        transpose = opset18.Transpose(view_3, perm=[0, 2, 1])
        cat = opset18.Concat(transpose, transpose, axis=-1)
        cos = opset18.Cos(cat)
        sin = opset18.Sin(cat)
        mul_1 = cos * 1.0
        mul_2 = sin * 1.0
        _to_copy_4 = opset18.Cast(mul_1, to=1)
        _to_copy_5 = opset18.Cast(mul_2, to=1)
        _to_copy_6 = opset18.Cast(embedding, to=1)
        scalar_tensor_default = opset18.Cast(2, to=1)
        pow_1 = _to_copy_6**scalar_tensor_default
        val_55 = opset18.Constant(value_ints=[-1])
        val_57 = opset18.Reshape([-1], val_55, allowzero=0)
        mean = opset18.ReduceMean(pow_1, val_57, keepdims=1, noop_with_empty_axes=0)
        add = mean + 1e-05
        val_59 = opset18.Sqrt(add)
        rsqrt = opset18.Reciprocal(val_59)
        mul_3 = _to_copy_6 * rsqrt
        _to_copy_7 = opset18.Cast(mul_3, to=1)
        mul_4 = model_layers_0_input_layernorm_weight * _to_copy_7
        t = opset18.Transpose(model_layers_0_self_attn_q_proj_weight, perm=[1, 0])
        val_61 = opset18.Cast([30, 2048], to=7)
        view_4 = opset18.Reshape(mul_4, val_61, allowzero=0)
        mm = view_4 @ t
        val_63 = opset18.Cast([1, 30, 2048], to=7)
        view_5 = opset18.Reshape(mm, val_63, allowzero=0)
        t_1 = opset18.Transpose(model_layers_0_self_attn_k_proj_weight, perm=[1, 0])
        val_64 = opset18.Cast([30, 2048], to=7)
        view_6 = opset18.Reshape(mul_4, val_64, allowzero=0)
        mm_1 = view_6 @ t_1
        val_65 = opset18.Cast([1, 30, 2048], to=7)
        view_7 = opset18.Reshape(mm_1, val_65, allowzero=0)
        t_2 = opset18.Transpose(model_layers_0_self_attn_v_proj_weight, perm=[1, 0])
        val_66 = opset18.Cast([30, 2048], to=7)
        view_8 = opset18.Reshape(mul_4, val_66, allowzero=0)
        mm_2 = view_8 @ t_2
        val_67 = opset18.Cast([1, 30, 2048], to=7)
        view_9 = opset18.Reshape(mm_2, val_67, allowzero=0)
        val_69 = opset18.Cast([1, 30, 32, 64], to=7)
        view_10 = opset18.Reshape(view_5, val_69, allowzero=0)
        transpose_1 = opset18.Transpose(view_10, perm=[0, 2, 1, 3])
        val_70 = opset18.Cast([1, 30, 32, 64], to=7)
        view_11 = opset18.Reshape(view_7, val_70, allowzero=0)
        transpose_2 = opset18.Transpose(view_11, perm=[0, 2, 1, 3])
        val_71 = opset18.Cast([1, 30, 32, 64], to=7)
        view_12 = opset18.Reshape(view_9, val_71, allowzero=0)
        transpose_3 = opset18.Transpose(view_12, perm=[0, 2, 1, 3])
        dim__8 = opset18.Constant(value_ints=[1])
        dim_0__8 = opset18.Cast(dim__8, to=7)
        unsqueeze_3 = opset18.Unsqueeze(_to_copy_4, dim_0__8)
        dim__9 = opset18.Constant(value_ints=[1])
        dim_0__9 = opset18.Cast(dim__9, to=7)
        unsqueeze_4 = opset18.Unsqueeze(_to_copy_5, dim_0__9)
        mul_5 = transpose_1 * unsqueeze_3
        val_72 = opset18.Cast(0, to=7)
        val_73 = opset18.Constant(value_ints=[-1])
        val_74 = opset18.Reshape(val_72, val_73, allowzero=0)
        val_76 = opset18.Cast(32, to=7)
        val_77 = opset18.Constant(value_ints=[-1])
        val_78 = opset18.Reshape(val_76, val_77, allowzero=0)
        val_80 = opset18.Cast(3, to=7)
        val_81 = opset18.Constant(value_ints=[-1])
        val_82 = opset18.Reshape(val_80, val_81, allowzero=0)
        val_83 = opset18.Constant(value_ints=[1])
        slice_4 = opset18.Slice(transpose_1, val_74, val_78, val_82, val_83)
        val_84 = opset18.Cast(32, to=7)
        val_85 = opset18.Constant(value_ints=[-1])
        val_86 = opset18.Reshape(val_84, val_85, allowzero=0)
        val_87 = opset18.Cast(9223372036854775807, to=7)
        val_88 = opset18.Constant(value_ints=[-1])
        val_89 = opset18.Reshape(val_87, val_88, allowzero=0)
        val_90 = opset18.Cast(3, to=7)
        val_91 = opset18.Constant(value_ints=[-1])
        val_92 = opset18.Reshape(val_90, val_91, allowzero=0)
        val_93 = opset18.Constant(value_ints=[1])
        slice_5 = opset18.Slice(transpose_1, val_86, val_89, val_92, val_93)
        neg = opset18.Neg(slice_5)
        cat_1 = opset18.Concat(neg, slice_4, axis=-1)
        mul_6 = cat_1 * unsqueeze_4
        add_1 = mul_5 + mul_6
        mul_7 = transpose_2 * unsqueeze_3
        val_94 = opset18.Cast(0, to=7)
        val_95 = opset18.Constant(value_ints=[-1])
        val_96 = opset18.Reshape(val_94, val_95, allowzero=0)
        val_97 = opset18.Cast(32, to=7)
        val_98 = opset18.Constant(value_ints=[-1])
        val_99 = opset18.Reshape(val_97, val_98, allowzero=0)
        val_100 = opset18.Cast(3, to=7)
        val_101 = opset18.Constant(value_ints=[-1])
        val_102 = opset18.Reshape(val_100, val_101, allowzero=0)
        val_103 = opset18.Constant(value_ints=[1])
        slice_6 = opset18.Slice(transpose_2, val_96, val_99, val_102, val_103)
        val_104 = opset18.Cast(32, to=7)
        val_105 = opset18.Constant(value_ints=[-1])
        val_106 = opset18.Reshape(val_104, val_105, allowzero=0)
        val_107 = opset18.Cast(9223372036854775807, to=7)
        val_108 = opset18.Constant(value_ints=[-1])
        val_109 = opset18.Reshape(val_107, val_108, allowzero=0)
        val_110 = opset18.Cast(3, to=7)
        val_111 = opset18.Constant(value_ints=[-1])
        val_112 = opset18.Reshape(val_110, val_111, allowzero=0)
        val_113 = opset18.Constant(value_ints=[1])
        slice_7 = opset18.Slice(transpose_2, val_106, val_109, val_112, val_113)
        neg_1 = opset18.Neg(slice_7)
        cat_2 = opset18.Concat(neg_1, slice_6, axis=-1)
        mul_8 = cat_2 * unsqueeze_4
        add_2 = mul_7 + mul_8
        cat_3 = opset18.Concat(past_key_values_0_0, add_2, axis=-2)
        cat_4 = opset18.Concat(past_key_values_0_1, transpose_3, axis=-2)
        dim__10 = opset18.Constant(value_ints=[0])
        dim_0__10 = opset18.Cast(dim__10, to=7)
        unsqueeze_5 = opset18.Unsqueeze(mul, dim_0__10)
        dim__11 = opset18.Constant(value_ints=[1])
        dim_0__11 = opset18.Cast(dim__11, to=7)
        unsqueeze_6 = opset18.Unsqueeze(unsqueeze_5, dim_0__11)
        val_114 = opset18.Cast(0, to=7)
        val_115 = opset18.Constant(value_ints=[-1])
        val_116 = opset18.Reshape(val_114, val_115, allowzero=0)
        val_117 = opset18.Cast(9223372036854775807, to=7)
        val_118 = opset18.Constant(value_ints=[-1])
        val_119 = opset18.Reshape(val_117, val_118, allowzero=0)
        val_120 = opset18.Cast(2, to=7)
        val_121 = opset18.Constant(value_ints=[-1])
        val_122 = opset18.Reshape(val_120, val_121, allowzero=0)
        val_123 = opset18.Constant(value_ints=[1])
        slice_8 = opset18.Slice(unsqueeze_6, val_116, val_119, val_122, val_123)
        val_124 = opset18.Cast(0, to=7)
        val_125 = opset18.Constant(value_ints=[-1])
        val_126 = opset18.Reshape(val_124, val_125, allowzero=0)
        val_127 = opset18.Cast(9223372036854775807, to=7)
        val_128 = opset18.Constant(value_ints=[-1])
        val_129 = opset18.Reshape(val_127, val_128, allowzero=0)
        val_130 = opset18.Cast(3, to=7)
        val_131 = opset18.Constant(value_ints=[-1])
        val_132 = opset18.Reshape(val_130, val_131, allowzero=0)
        val_133 = opset18.Constant(value_ints=[1])
        slice_9 = opset18.Slice(slice_8, val_126, val_129, val_132, val_133)
        size_0__12 = opset18.Cast([1, 1, -1, -1], to=7)
        size_1__12 = opset18.Abs(size_0__12)
        expand_3 = opset18.Expand(slice_9, size_1__12)
        val_135 = opset18.Cast(0, to=7)
        val_136 = opset18.Constant(value_ints=[-1])
        val_137 = opset18.Reshape(val_135, val_136, allowzero=0)
        val_138 = opset18.Cast(9223372036854775807, to=7)
        val_139 = opset18.Constant(value_ints=[-1])
        val_140 = opset18.Reshape(val_138, val_139, allowzero=0)
        val_141 = opset18.Cast(0, to=7)
        val_142 = opset18.Constant(value_ints=[-1])
        val_143 = opset18.Reshape(val_141, val_142, allowzero=0)
        val_144 = opset18.Constant(value_ints=[1])
        slice_10 = opset18.Slice(expand_3, val_137, val_140, val_143, val_144)
        val_145 = opset18.Cast(0, to=7)
        val_146 = opset18.Constant(value_ints=[-1])
        val_147 = opset18.Reshape(val_145, val_146, allowzero=0)
        val_148 = opset18.Cast(9223372036854775807, to=7)
        val_149 = opset18.Constant(value_ints=[-1])
        val_150 = opset18.Reshape(val_148, val_149, allowzero=0)
        val_151 = opset18.Cast(1, to=7)
        val_152 = opset18.Constant(value_ints=[-1])
        val_153 = opset18.Reshape(val_151, val_152, allowzero=0)
        val_154 = opset18.Constant(value_ints=[1])
        slice_11 = opset18.Slice(slice_10, val_147, val_150, val_153, val_154)
        val_155 = opset18.Cast(0, to=7)
        val_156 = opset18.Constant(value_ints=[-1])
        val_157 = opset18.Reshape(val_155, val_156, allowzero=0)
        val_158 = opset18.Cast(9223372036854775807, to=7)
        val_159 = opset18.Constant(value_ints=[-1])
        val_160 = opset18.Reshape(val_158, val_159, allowzero=0)
        val_161 = opset18.Cast(2, to=7)
        val_162 = opset18.Constant(value_ints=[-1])
        val_163 = opset18.Reshape(val_161, val_162, allowzero=0)
        val_164 = opset18.Constant(value_ints=[1])
        slice_12 = opset18.Slice(slice_11, val_157, val_160, val_163, val_164)
        val_165 = opset18.Cast(0, to=7)
        val_166 = opset18.Constant(value_ints=[-1])
        val_167 = opset18.Reshape(val_165, val_166, allowzero=0)
        val_168 = opset18.Cast(46, to=7)
        val_169 = opset18.Constant(value_ints=[-1])
        val_170 = opset18.Reshape(val_168, val_169, allowzero=0)
        val_171 = opset18.Cast(3, to=7)
        val_172 = opset18.Constant(value_ints=[-1])
        val_173 = opset18.Reshape(val_171, val_172, allowzero=0)
        val_174 = opset18.Constant(value_ints=[1])
        slice_13 = opset18.Slice(slice_12, val_167, val_170, val_173, val_174)
        val_175 = opset18.Shape(add_1, start=0)
        val_176 = opset18.Constant(value_ints=[-1])
        val_177 = opset18.Gather(val_175, val_176, axis=0)
        val_178 = opset18.CastLike(val_177, add_1)
        val_179 = opset18.Constant(value_float=1.0)
        val_180 = opset18.CastLike(val_179, add_1)
        val_181 = opset18.Sqrt(val_178)
        val_182 = val_180 / val_181
        val_183 = opset18.CastLike(val_182, add_1)
        val_184 = opset18.Shape(cat_3, start=0)
        val_185 = opset18.Constant(value_ints=[9223372036854775807])
        val_186 = opset18.Slice(val_184, [-1], val_185)
        val_188 = opset18.Slice(val_184, [-2], [-1])
        val_189 = opset18.Constant(value_ints=[-9223372036854775808])
        val_190 = opset18.Slice(val_184, val_189, [-2])
        val_191 = opset18.Constant(value_ints=[-1])
        val_192 = opset18.Concat(val_191, val_188, val_186, axis=0)
        val_193 = opset18.Reshape(cat_3, val_192, allowzero=0)
        val_194 = opset18.Transpose(val_193, perm=[0, 2, 1])
        val_195 = opset18.Concat(val_190, val_186, val_188, axis=0)
        val_196 = opset18.Reshape(val_194, val_195, allowzero=0)
        val_197 = opset18.Sqrt(val_183)
        val_198 = add_1 * val_197
        val_199 = opset18.Sqrt(val_183)
        val_200 = val_196 * val_199
        val_201 = val_198 @ val_200
        val_202 = val_201 + slice_13
        val_203 = opset18.Softmax(val_202, axis=-1)
        val_204, _unused = opset18.Dropout(val_203, 0.0)
        getitem = val_204 @ cat_4
        val_206 = opset18.Shape(add_1, start=0)
        val_209 = opset18.Slice(val_206, [0], [1])
        val_211 = opset18.Slice(val_206, [1], [2])
        val_212 = opset18.Slice(val_206, [-2], [-1])
        val_213 = opset18.Cast(val_211, to=1)
        val_215 = val_213 / 32.0
        val_216 = opset18.Ceil(val_215)
        val_217 = val_216 * 32.0
        val_218 = opset18.Cast(val_217, to=7)
        val_219 = opset18.Concat(val_209, val_212, val_218, axis=0)
        _scaled_dot_product_flash_attention_for_cpu__1 = opset18.Expand(0.0, val_219)
        transpose_4 = opset18.Transpose(getitem, perm=[0, 2, 1, 3])
        val_221 = opset18.Cast([1, 30, -1], to=7)
        view_13 = opset18.Reshape(transpose_4, val_221, allowzero=0)
        t_3 = opset18.Transpose(model_layers_0_self_attn_o_proj_weight, perm=[1, 0])
        val_222 = opset18.Cast([30, 2048], to=7)
        view_14 = opset18.Reshape(view_13, val_222, allowzero=0)
        mm_3 = view_14 @ t_3
        val_223 = opset18.Cast([1, 30, 2048], to=7)
        view_15 = opset18.Reshape(mm_3, val_223, allowzero=0)
        add_3 = embedding + view_15
        _to_copy_8 = opset18.Cast(add_3, to=1)
        scalar_tensor_default_1 = opset18.Cast(2, to=1)
        pow_2 = _to_copy_8**scalar_tensor_default_1
        val_224 = opset18.Constant(value_ints=[-1])
        val_225 = opset18.Reshape([-1], val_224, allowzero=0)
        mean_1 = opset18.ReduceMean(pow_2, val_225, keepdims=1, noop_with_empty_axes=0)
        add_4 = mean_1 + 1e-05
        val_226 = opset18.Sqrt(add_4)
        rsqrt_1 = opset18.Reciprocal(val_226)
        mul_9 = _to_copy_8 * rsqrt_1
        _to_copy_9 = opset18.Cast(mul_9, to=1)
        mul_10 = model_layers_0_post_attention_layernorm_weight * _to_copy_9
        t_4 = opset18.Transpose(model_layers_0_mlp_gate_proj_weight, perm=[1, 0])
        val_227 = opset18.Cast([30, 2048], to=7)
        view_16 = opset18.Reshape(mul_10, val_227, allowzero=0)
        mm_4 = view_16 @ t_4
        val_229 = opset18.Cast([1, 30, 8192], to=7)
        view_17 = opset18.Reshape(mm_4, val_229, allowzero=0)
        val_230 = opset18.Sigmoid(view_17)
        silu = view_17 * val_230
        t_5 = opset18.Transpose(model_layers_0_mlp_up_proj_weight, perm=[1, 0])
        val_231 = opset18.Cast([30, 2048], to=7)
        view_18 = opset18.Reshape(mul_10, val_231, allowzero=0)
        mm_5 = view_18 @ t_5
        val_232 = opset18.Cast([1, 30, 8192], to=7)
        view_19 = opset18.Reshape(mm_5, val_232, allowzero=0)
        mul_11 = silu * view_19
        t_6 = opset18.Transpose(model_layers_0_mlp_down_proj_weight, perm=[1, 0])
        val_234 = opset18.Cast([30, 8192], to=7)
        view_20 = opset18.Reshape(mul_11, val_234, allowzero=0)
        mm_6 = view_20 @ t_6
        val_235 = opset18.Cast([1, 30, 2048], to=7)
        view_21 = opset18.Reshape(mm_6, val_235, allowzero=0)
        add_5 = add_3 + view_21
        _to_copy_10 = opset18.Cast(add_5, to=1)
        scalar_tensor_default_2 = opset18.Cast(2, to=1)
        pow_3 = _to_copy_10**scalar_tensor_default_2
        val_236 = opset18.Constant(value_ints=[-1])
        val_237 = opset18.Reshape([-1], val_236, allowzero=0)
        mean_2 = opset18.ReduceMean(pow_3, val_237, keepdims=1, noop_with_empty_axes=0)
        add_6 = mean_2 + 1e-05
        val_238 = opset18.Sqrt(add_6)
        rsqrt_2 = opset18.Reciprocal(val_238)
        mul_12 = _to_copy_10 * rsqrt_2
        _to_copy_11 = opset18.Cast(mul_12, to=1)
        mul_13 = model_norm_weight * _to_copy_11
        t_7 = opset18.Transpose(lm_head_weight, perm=[1, 0])
        val_239 = opset18.Cast([30, 2048], to=7)
        view_22 = opset18.Reshape(mul_13, val_239, allowzero=0)
        mm_7 = view_22 @ t_7
        val_241 = opset18.Cast([1, 30, 49152], to=7)
        view_23 = opset18.Reshape(mm_7, val_241, allowzero=0)
        _to_copy_12 = opset18.Cast(view_23, to=1)
        return _to_copy_12, cat_3, cat_4

    model = main_graph.to_model_proto()
    return model


def make_model_with_random_weights():
    model_layers_0_input_layernorm_weight = numpy.random.rand(2048).astype(numpy.float32)
    model_layers_0_post_attention_layernorm_weight = numpy.random.rand(2048).astype(
        numpy.float32
    )
    model_norm_weight = numpy.random.rand(2048).astype(numpy.float32)
    lm_head_weight = numpy.random.rand(49152, 2048).astype(numpy.float32)
    model_layers_0_self_attn_q_proj_weight = numpy.random.rand(2048, 2048).astype(
        numpy.float32
    )
    model_layers_0_self_attn_k_proj_weight = numpy.random.rand(2048, 2048).astype(
        numpy.float32
    )
    model_layers_0_self_attn_v_proj_weight = numpy.random.rand(2048, 2048).astype(
        numpy.float32
    )
    model_layers_0_self_attn_o_proj_weight = numpy.random.rand(2048, 2048).astype(
        numpy.float32
    )
    model_layers_0_mlp_gate_proj_weight = numpy.random.rand(8192, 2048).astype(numpy.float32)
    model_layers_0_mlp_up_proj_weight = numpy.random.rand(8192, 2048).astype(numpy.float32)
    model_layers_0_mlp_down_proj_weight = numpy.random.rand(2048, 8192).astype(numpy.float32)
    model_rotary_emb_inv_freq = numpy.random.rand(32).astype(numpy.float32)
    model = make_model(
        model_layers_0_input_layernorm_weight,
        model_layers_0_post_attention_layernorm_weight,
        model_norm_weight,
        lm_head_weight,
        model_layers_0_self_attn_q_proj_weight,
        model_layers_0_self_attn_k_proj_weight,
        model_layers_0_self_attn_v_proj_weight,
        model_layers_0_self_attn_o_proj_weight,
        model_layers_0_mlp_gate_proj_weight,
        model_layers_0_mlp_up_proj_weight,
        model_layers_0_mlp_down_proj_weight,
        model_rotary_emb_inv_freq,
    )
    return model


class _SmollmTest2:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model

    def get_ort_inputs(self):
        if not hasattr(self, "_ort_inputs"):
            inputs = {
                "input_ids": numpy.random.randint(0, 49152, (1, 30)).astype(numpy.int64),
                "position_ids": numpy.arange(30).reshape(1, 30).astype(numpy.int64),
                "past_key_values_0_0": numpy.random.rand(1, 32, 16, 64).astype(numpy.float32),
                "past_key_values_0_1": numpy.random.rand(1, 32, 16, 64).astype(numpy.float32),
            }
            self._ort_inputs = inputs
        return self._ort_inputs


def smollm_test_2():
    return _SmollmTest2()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
# pylint: disable=not-callable

import sys
import unittest

import numpy as np
import onnxruntime
import torch

import onnxscript.tools.transformers_models
import onnxscript.tools.transformers_models.phi3
from onnxscript._internal.version_utils import (
    has_transformers,
    ignore_warnings,
    torch_older_than,
)

has_phi3 = onnxscript.tools.transformers_models.phi3.has_phi3


class TestExportPhi3(unittest.TestCase):
    @unittest.skipIf(sys.platform == "win32", reason="not supported yet on Windows")
    @unittest.skipIf(not has_transformers(), reason="transformers is missing")
    @unittest.skipIf(not has_phi3(), reason="transformers is not recent enough")
    @unittest.skipIf(torch_older_than("2.4"), reason="fails to export")
    @ignore_warnings(UserWarning)
    def test_phi3_export_cpu(self):
        model, input_tensors_many, _ = (
            onnxscript.tools.transformers_models.phi3.get_phi3_model()
        )
        input_tensors = input_tensors_many[0]
        expected = model(*input_tensors)
        proto = onnxscript.tools.transformers_models.export_to_onnx(model, *input_tensors)
        names = [i.name for i in proto.graph.input]
        np_input_tensors = [x.numpy() for x in input_tensors]
        feeds = dict(zip(names, np_input_tensors))
        sess = onnxruntime.InferenceSession(
            proto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        results = sess.run(None, feeds)
        np.testing.assert_allclose(expected[0].detach().numpy(), results[0], atol=1e-5)

    @unittest.skipIf(sys.platform == "win32", reason="not supported yet on Windows")
    @unittest.skipIf(not has_transformers(), reason="transformers is missing")
    @unittest.skipIf(not has_phi3(), reason="transformers is not recent enough")
    @unittest.skipIf(torch_older_than("2.5"), reason="fails to export")
    @ignore_warnings(UserWarning)
    def test_phi3_export_cpu_export_api(self):
        model, input_tensors_many, _ = (
            onnxscript.tools.transformers_models.phi3.get_phi3_model()
        )
        input_tensors = input_tensors_many[0]
        expected = model(*input_tensors)
        proto = onnxscript.tools.transformers_models.export_to_onnx(
            model, *input_tensors, export_api=True
        )
        names = [i.name for i in proto.graph.input]
        np_input_tensors = [x.numpy() for x in input_tensors]
        feeds = dict(zip(names, np_input_tensors))
        sess = onnxruntime.InferenceSession(
            proto.SerializeToString(), providers=["CPUExecutionProvider"]
        )
        results = sess.run(None, feeds)
        np.testing.assert_allclose(expected[0].detach().numpy(), results[0], atol=1e-5)

    @unittest.skipIf(sys.platform == "win32", reason="not supported yet on Windows")
    @unittest.skipIf(not torch.cuda.is_available(), reason="CUDA not available")
    @unittest.skipIf(not has_transformers(), reason="transformers is missing")
    @unittest.skipIf(not has_phi3(), reason="transformers is not recent enough")
    @ignore_warnings(UserWarning)
    def test_phi3_export_cuda(self):
        model, input_tensors_many, _ = (
            onnxscript.tools.transformers_models.phi3.get_phi3_model()
        )
        input_tensors_cpu = input_tensors_many[0]
        model = model.to("cuda")
        input_tensors = [i.to("cuda") for i in input_tensors_cpu]
        expected = model(*input_tensors)
        proto = onnxscript.tools.transformers_models.export_to_onnx(model, *input_tensors)
        names = [i.name for i in proto.graph.input]
        np_input_tensors = [x.detach().cpu().numpy() for x in input_tensors]
        feeds = dict(zip(names, np_input_tensors))
        sess = onnxruntime.InferenceSession(
            proto.SerializeToString(), providers=["CUDAExecutionProvider"]
        )
        results = sess.run(None, feeds)
        np.testing.assert_allclose(expected[0].detach().cpu().numpy(), results[0], atol=1e-5)


if __name__ == "__main__":
    unittest.main(verbosity=2)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import math
import unittest

import numpy as np
import onnx
import onnx_ir as ir
import onnx_ir.passes.common.shape_inference as shape_inference
import onnxruntime as ort
import parameterized
import torch

import onnxscript
import onnxscript.optimizer
from onnxscript import FLOAT, script
from onnxscript import opset18 as op
from onnxscript.rewriter.models._phi4lm import phi4lm_test
from onnxscript.rewriter.ort_fusions import optimize_for_ort
from onnxscript.rewriter.ort_fusions._test_utils import assert_allclose
from onnxscript.rewriter.ort_fusions.gqa import fuse_gqa
from onnxscript.rewriter.ort_fusions.sdpa import fuse_sdpa

msft_op = onnxscript.values.Opset("com.microsoft", 1)

# Test case for GroupQueryAttention (GQA) fusion.


class GQAFusionTest(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Config parameters
        self.batchsize = 1  # Note: GQA (cpu) seems to require batch-size 1?
        self.seqlen = 8
        self.kv_seqlen = self.seqlen
        self.past_seqlen = 16
        self.head_size = 16
        self.num_heads = 20
        self.kv_num_heads = 10

        # Computed config parameters
        self.hidden_size = self.head_size * self.num_heads
        self.kv_hidden_size = self.head_size * self.kv_num_heads
        assert (self.num_heads % self.kv_num_heads) == 0, (
            "num_heads must be divisible by kv_num_heads"
        )
        self.num_groups = self.num_heads // self.kv_num_heads
        self.total_seqlen = self.seqlen + self.past_seqlen

        # Abbreviations
        B = self.batchsize
        S = self.seqlen
        P = self.past_seqlen
        D = self.hidden_size
        Dkv = self.kv_hidden_size
        Dh = self.head_size
        Hkv = self.kv_num_heads
        total_seqlen = S + P
        max_seqlen = total_seqlen

        # Input/output types have some dimensions as dynamic (even though the
        # test case instance has specific values above).
        self.input_types = (
            FLOAT["B", "S", D],  # query
            FLOAT["B", "S", Dkv],  # key
            FLOAT["B", "S", Dkv],  # value
            FLOAT["B", Hkv, "P", Dh],  # past_key
            FLOAT["B", Hkv, "P", Dh],  # past_value
            FLOAT["max_seqlen", Dh // 2],  # cos
            FLOAT["max_seqlen", Dh // 2],  # sin
        )
        self.output_types = (
            FLOAT["B", "S", D],  # attention
            FLOAT["B", Hkv, "T", Dh],  # present_key
            FLOAT["B", Hkv, "T", Dh],  # present_value
        )

        self.inputs = {
            "query": np.random.rand(B, S, D).astype(np.float32),
            "key": np.random.rand(B, S, Dkv).astype(np.float32),
            "value": np.random.rand(B, S, Dkv).astype(np.float32),
            "past_key": np.random.rand(B, Hkv, P, Dh).astype(np.float32),
            "past_value": np.random.rand(B, Hkv, P, Dh).astype(np.float32),
            "cos": np.random.rand(max_seqlen, Dh // 2).astype(np.float32),
            "sin": np.random.rand(max_seqlen, Dh // 2).astype(np.float32),
        }

    def target_model_script(self):
        H = self.num_heads
        Hkv = self.kv_num_heads

        @script()
        def gqa(query, key, value, past_key, past_value, cos, sin):
            # Generate seqlens_k and total_seqlen inputs for GQA:
            # In this test case, all batch elements have same sequence length.
            S = op.Shape(query, start=1, end=2)
            past_seq_length = op.Shape(past_key, start=2, end=3)
            total_seq_length = op.Add(past_seq_length, S)
            total_seqlen_int32 = op.Cast(total_seq_length, to=6)
            total_seqlen_int32_minus_1 = op.Sub(total_seqlen_int32, 1)
            batchsize = op.Shape(query, start=0, end=1)
            seqlens_k = op.Tile(total_seqlen_int32_minus_1, batchsize)

            attn, past_key, past_value = msft_op.GroupQueryAttention(
                query,
                key,
                value,
                past_key,
                past_value,
                seqlens_k,
                total_seqlen_int32,
                cos,
                sin,
                num_heads=H,
                kv_num_heads=Hkv,
                do_rotary=1,
            )
            return attn, past_key, past_value

        return gqa

    def source_model_script(self):
        scale_factor = math.sqrt(math.sqrt(self.head_size))
        minval = torch.finfo(torch.float32).min
        minval_tp = onnx.helper.make_tensor("minval", onnx.TensorProto.FLOAT, [1], [minval])
        H = [self.num_heads]
        Hkv = [self.kv_num_heads]
        Dh = [self.head_size]
        G = [self.num_groups]
        minus_1 = [-1]  # inferred dimension in Reshape op
        plus_1 = [1]

        @script()
        def gqa(query, key, value, past_key, past_value, cos, sin):
            # Shapes used for Reshape ops. Note that we have a few different options on how shapes are
            # specified in an ONNX Reshape op (which supports special values 0 and -1 to propagate
            # existing dimension and one inferred dimension respectively). The following shapes are
            # based on what is observed in Phi models generated by the exporter.
            B = op.Shape(query, start=0, end=1)
            S = op.Shape(query, start=1, end=2)
            past_seq_length = op.Shape(past_key, start=2, end=3)
            total_seq_length = op.Add(past_seq_length, S)
            # past_seq_length = op.Squeeze(past_seq_length_1D, [0])
            # S_0D = op.Squeeze(S,[0])

            shape_BSHDh = op.Concat(B, S, minus_1, Dh, axis=0)
            shape_BSHkvDh = op.Concat(B, S, minus_1, Dh, axis=0)
            shape_BSD = op.Concat(B, S, minus_1, axis=0)
            shape_BHkvGSDh = op.Concat(B, Hkv, G, total_seq_length, Dh, axis=0)

            shape_BHSDh = op.Concat(B, H, total_seq_length, Dh, axis=0)

            # First, get Q, K, V into right shapes. Inputs are 3D tensors in the BSD format.
            # D is different for Q and K/V (not reflected in the names, unfortunately).
            # We convert them into BHSDh (i.e., BHSd) format. In this version, we have only
            # one sequence length (S) for all Q, K, and V (with no cache).
            query_BSHDh = op.Reshape(query, shape_BSHDh)
            query_BHSDh = op.Transpose(query_BSHDh, perm=[0, 2, 1, 3])

            key_BSHkvDh = op.Reshape(key, shape_BSHkvDh)
            key_BHkvSDh = op.Transpose(key_BSHkvDh, perm=[0, 2, 1, 3])

            value_BSHkvDh = op.Reshape(value, shape_BSHkvDh)
            value_BHkvSDh = op.Transpose(value_BSHkvDh, perm=[0, 2, 1, 3])

            # Concat past and do rotary embedding
            position_ids_1d = op.Range(past_seq_length, total_seq_length, 1)
            position_ids_q = op.Unsqueeze(position_ids_1d, [0])
            position_ids_k = op.Unsqueeze(position_ids_1d, [0])

            # Note: The above code pattern for position-ids is from exported Phi model.
            # However, for use with ORT's RotaryEmbedding it needs the following for batchsize > 1
            # But we currently target batchsize=1 since GQA requires it when there is a past key/value.
            #
            # position_ids_2d = op.Unsqueeze(position_ids_1d, [0])
            # tile_B_1 = op.Concat(B, plus_1, axis=0)
            # position_ids = op.Tile(position_ids_2d, tile_B_1)

            query_BHSDh_rope = msft_op.RotaryEmbedding(
                query_BHSDh,
                position_ids_q,
                cos,
                sin,
            )
            key_BHkvSDh_rope = msft_op.RotaryEmbedding(
                key_BHkvSDh,
                position_ids_k,
                cos,
                sin,
            )
            key_seq_BHkvSkvDh = op.Concat(past_key, key_BHkvSDh_rope, axis=-2)

            value_seq_BHkvSkvDh = op.Concat(past_value, value_BHkvSDh, axis=-2)

            # Now, expand from shared heads to all heads
            key_BHkv1SDh = op.Unsqueeze(key_seq_BHkvSkvDh, [2])
            key_BHkvGSDh = op.Expand(key_BHkv1SDh, shape_BHkvGSDh)
            key_BHSDh = op.Reshape(key_BHkvGSDh, shape_BHSDh)

            value_BHkv1SDh = op.Unsqueeze(value_seq_BHkvSkvDh, [2])
            value_BHkvGSDh = op.Expand(value_BHkv1SDh, shape_BHkvGSDh)
            value_BHSDh = op.Reshape(value_BHkvGSDh, shape_BHSDh)

            # Generate causal mask:
            # where every row looks like [0, 0, ..., /*diagonal=*/ 0, minval, minval, ...]
            seq_len = op.Shape(query, end=2, start=1)
            seq_len_0D = op.Squeeze(seq_len)

            past_seq_len_0D = op.Squeeze(past_seq_length)

            total_seq_len_0D = op.Add(past_seq_len_0D, seq_len_0D)
            total_seq_len = op.Reshape(total_seq_len_0D, [-1])

            # The Phi modeling code generates the following +1 as the target-length, which seems
            # unnecessary in this context. But duplicating same logic here.
            total_seq_len_plus_1_0D = op.Add(total_seq_len_0D, 1)
            total_seq_len_plus_1 = op.Reshape(total_seq_len_plus_1_0D, [-1])

            current_range = op.Range(past_seq_len_0D, total_seq_len_0D, 1)
            mask_shape = op.Concat(seq_len, total_seq_len_plus_1, axis=0)
            min_val = op.Constant(value=minval_tp)
            mask_all_min = op.Expand(min_val, mask_shape)
            total_range_as_row = op.Range(0, total_seq_len_plus_1_0D, 1)
            current_range_as_column = op.Reshape(current_range, [-1, 1])
            boolean_mask = op.Greater(total_range_as_row, current_range_as_column)
            float_0_1_mask = op.Cast(boolean_mask, to=1)
            float_0_min_mask = op.Mul(mask_all_min, float_0_1_mask)
            mask_4d = op.Unsqueeze(float_0_min_mask, [0, 1])
            shape_B111 = op.Concat(B, plus_1, plus_1, plus_1, axis=0)
            mask_B1ST_plus = op.Expand(mask_4d, shape_B111)

            # Get rid of the extra +1 added above: total_seq_len is enough, no
            # need for total_seq_len+1.
            mask_B1ST = op.Slice(mask_B1ST_plus, [0], total_seq_len, [3], [1])

            # Now, compute attention:
            key_transposed = op.Transpose(key_BHSDh, perm=[0, 1, 3, 2])
            divisor = op.Constant(value_float=scale_factor)
            scaled_query = op.Div(query_BHSDh_rope, divisor)
            scaled_key = op.Div(key_transposed, divisor)
            attn_score = op.MatMul(scaled_query, scaled_key)
            masked_attn_score = op.Add(attn_score, mask_B1ST)
            attn_weight = op.Softmax(masked_attn_score, axis=-1)
            attention_BHSDh = op.MatMul(attn_weight, value_BHSDh)

            # Reshape back to BSD format
            attention_BSHDh = op.Transpose(attention_BHSDh, perm=[0, 2, 1, 3])
            attention_BSD = op.Reshape(attention_BSHDh, shape_BSD)

            return attention_BSD, key_seq_BHkvSkvDh, value_seq_BHkvSkvDh

        return gqa

    def test_equivalence(self):
        """Test that the source and target models produce the same outputs."""
        inputs = self.inputs

        source_model = self.source_model_script().to_model_proto(
            input_types=self.input_types,
            output_types=self.output_types,
        )
        session = ort.InferenceSession(
            source_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        source_model_outputs = session.run(None, inputs)

        target_model = self.target_model_script().to_model_proto(
            input_types=self.input_types,
            output_types=self.output_types,
        )
        session = ort.InferenceSession(
            target_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        target_model_outputs = session.run(None, inputs)

        self.assertEqual(len(source_model_outputs), len(target_model_outputs))
        assert_allclose(source_model_outputs, target_model_outputs)

    def test_fusion(self):
        """Test that GQA fusion is successful on source model and produces an equivalent model."""
        inputs = self.inputs

        source_model = self.source_model_script().to_model_proto(
            input_types=self.input_types,
            output_types=self.output_types,
        )
        session = ort.InferenceSession(
            source_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        source_model_outputs = session.run(None, inputs)

        # Some shapes need to be present in input model for fusion to be successful.
        # (i) Shape inference doesn't handle handle ORT contrib ops.
        # (ii) TODO: investigate if Reshape(..., ["B", "S", -1, Dh]) handled precisely
        # by shape inference.
        query_BHSDh_rope_value_info = onnx.helper.make_tensor_value_info(
            "query_BHSDh_rope",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.seqlen, self.head_size],
        )
        key_BHkvSDh_rope_value_info = onnx.helper.make_tensor_value_info(
            "key_BHkvSDh_rope",
            onnx.TensorProto.FLOAT,
            ["B", self.kv_num_heads, self.seqlen, self.head_size],
        )
        query_BSHDh_value_info = onnx.helper.make_tensor_value_info(
            "query_BSHDh",
            onnx.TensorProto.FLOAT,
            ["B", self.seqlen, self.num_heads, self.head_size],
        )
        key_BHSDh_value_info = onnx.helper.make_tensor_value_info(
            "key_BHSDh",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.total_seqlen, self.head_size],
        )
        key_BSHkvDh_value_info = onnx.helper.make_tensor_value_info(
            "key_BSHkvDh",
            onnx.TensorProto.FLOAT,
            ["B", self.seqlen, self.kv_num_heads, self.head_size],
        )
        key_transposed_value_info = onnx.helper.make_tensor_value_info(
            "key_transposed",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.head_size, self.total_seqlen],
        )
        value_BHSDh_value_info = onnx.helper.make_tensor_value_info(
            "value_BHSDh",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.total_seqlen, self.head_size],
        )
        source_model.graph.value_info.extend(
            [
                query_BHSDh_rope_value_info,
                key_BHkvSDh_rope_value_info,
                query_BSHDh_value_info,
                key_BHSDh_value_info,
                key_BSHkvDh_value_info,
                key_transposed_value_info,
                value_BHSDh_value_info,
            ]
        )

        source_model_ir = ir.serde.from_proto(source_model)
        inferred_model = shape_inference.infer_shapes(source_model_ir)
        onnxscript.optimizer.optimize(inferred_model)

        count = fuse_sdpa(inferred_model, debug=True)
        self.assertGreater(count, 0)

        count = fuse_gqa(inferred_model, debug=True)
        self.assertGreater(count, 0)

        fused_model = ir.serde.to_proto(inferred_model)
        session = ort.InferenceSession(
            fused_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        outputs3 = session.run(None, inputs)

        self.assertEqual(len(outputs3), len(source_model_outputs))
        assert_allclose(outputs3, source_model_outputs)


@parameterized.parameterized_class(
    [
        {"with_past": True, "transpose_first": True},
        {"with_past": True, "transpose_first": False},
        {"with_past": False, "transpose_first": True},
        {"with_past": False, "transpose_first": False},
    ]
)
class GemmaGQAFusionTest(unittest.TestCase):
    with_past = True
    transpose_first = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Config parameters
        self.batchsize = 1  # Note: GQA (cpu) seems to require batch-size 1?
        self.seqlen = 8
        self.kv_seqlen = self.seqlen
        self.past_seqlen = 16 if self.with_past else 0
        self.head_size = 16
        self.num_heads = 20
        self.kv_num_heads = 10

        # Computed config parameters
        self.hidden_size = self.head_size * self.num_heads
        self.kv_hidden_size = self.head_size * self.kv_num_heads
        assert (self.num_heads % self.kv_num_heads) == 0, (
            "num_heads must be divisible by kv_num_heads"
        )
        self.num_groups = self.num_heads // self.kv_num_heads
        self.total_seqlen = self.seqlen + self.past_seqlen

        # Abbreviations
        B = self.batchsize
        S = self.seqlen
        P = self.past_seqlen
        D = self.hidden_size
        Dkv = self.kv_hidden_size
        Dh = self.head_size
        Hkv = self.kv_num_heads
        total_seqlen = S + P
        max_seqlen = total_seqlen

        # Input/output types have some dimensions as dynamic (even though the
        # test case instance has specific values above).
        self.input_types = (
            FLOAT["B", "S", D],  # query
            FLOAT["B", "S", Dkv],  # key
            FLOAT["B", "S", Dkv],  # value
            FLOAT["B", Hkv, "P", Dh],  # past_key
            FLOAT["B", Hkv, "P", Dh],  # past_value
            FLOAT["max_seqlen", Dh // 2],  # cos
            FLOAT["max_seqlen", Dh // 2],  # sin
            FLOAT["Dh"],  # query_scale
            FLOAT["Dh"],  # key_scale
        )
        self.output_types = (
            FLOAT["B", "S", D],  # attention
            FLOAT["B", Hkv, "T", Dh],  # present_key
            FLOAT["B", Hkv, "T", Dh],  # present_value
        )

        self.inputs = {
            "query": np.random.rand(B, S, D).astype(np.float32),
            "key": np.random.rand(B, S, Dkv).astype(np.float32),
            "value": np.random.rand(B, S, Dkv).astype(np.float32),
            "past_key": np.random.rand(B, Hkv, P, Dh).astype(np.float32),
            "past_value": np.random.rand(B, Hkv, P, Dh).astype(np.float32),
            "cos": np.random.rand(max_seqlen, Dh // 2).astype(np.float32),
            "sin": np.random.rand(max_seqlen, Dh // 2).astype(np.float32),
            "query_scale": np.random.rand(Dh).astype(np.float32),
            "key_scale": np.random.rand(Dh).astype(np.float32),
        }

    def source_model_script(self):
        with_past = self.with_past
        transpose_first = self.transpose_first
        scale_factor = math.sqrt(math.sqrt(self.head_size))
        minval = torch.finfo(torch.float32).min
        minval_tp = onnx.helper.make_tensor("minval", onnx.TensorProto.FLOAT, [1], [minval])
        H = [self.num_heads]
        Hkv = [self.kv_num_heads]
        Dh = [self.head_size]
        G = [self.num_groups]
        minus_1 = [-1]  # inferred dimension in Reshape op
        plus_1 = [1]

        @script()
        def gqa(query, key, value, past_key, past_value, cos, sin, query_scale, key_scale):
            # Shapes used for Reshape ops. Note that we have a few different options on how shapes are
            # specified in an ONNX Reshape op (which supports special values 0 and -1 to propagate
            # existing dimension and one inferred dimension respectively). The following shapes are
            # based on what is observed in Phi models generated by the exporter.
            B = op.Shape(query, start=0, end=1)
            S = op.Shape(query, start=1, end=2)
            past_seq_length = op.Shape(past_key, start=2, end=3)
            total_seq_length = op.Add(past_seq_length, S)

            shape_BSHDh = op.Concat(B, S, minus_1, Dh, axis=0)
            shape_BSHkvDh = op.Concat(B, S, minus_1, Dh, axis=0)
            shape_BSD = op.Concat(B, S, minus_1, axis=0)
            shape_BHkvGSDh = op.Concat(B, Hkv, G, total_seq_length, Dh, axis=0)

            shape_BHSDh = op.Concat(B, H, total_seq_length, Dh, axis=0)

            # First, get Q, K, V into right shapes. Inputs are 3D tensors in the BSD format.
            # D is different for Q and K/V (not reflected in the names, unfortunately).
            # We convert them into BHSDh (i.e., BHSd) format. In this version, we have only
            # one sequence length (S) for all Q, K, and V (with no cache).
            query_BSHDh = op.Reshape(query, shape_BSHDh)
            key_BSHkvDh = op.Reshape(key, shape_BSHkvDh)

            if transpose_first:
                query_BHSDh = op.Transpose(query_BSHDh, perm=[0, 2, 1, 3])
                query_BHSDh_normalized = op.SimplifiedLayerNormalization(
                    query_BHSDh, query_scale, axis=-1, epsilon=1e-06, stash_type=1
                )
                key_BHkvSDh = op.Transpose(key_BSHkvDh, perm=[0, 2, 1, 3])
                key_BHkvSDh_normalized = op.SimplifiedLayerNormalization(
                    key_BHkvSDh, key_scale, axis=-1, epsilon=1e-06, stash_type=1
                )
            else:
                query_BSHDh_normalized = op.SimplifiedLayerNormalization(
                    query_BSHDh, query_scale, axis=-1, epsilon=1e-06, stash_type=1
                )
                query_BHSDh_normalized = op.Transpose(
                    query_BSHDh_normalized, perm=[0, 2, 1, 3]
                )
                key_BSHkvDh_normalized = op.SimplifiedLayerNormalization(
                    key_BSHkvDh, key_scale, axis=-1, epsilon=1e-06, stash_type=1
                )
                key_BHkvSDh_normalized = op.Transpose(
                    key_BSHkvDh_normalized, perm=[0, 2, 1, 3]
                )

            value_BSHkvDh = op.Reshape(value, shape_BSHkvDh)
            value_BHkvSDh = op.Transpose(value_BSHkvDh, perm=[0, 2, 1, 3])

            # Concat past and do rotary embedding
            position_ids_1d = op.Range(past_seq_length, total_seq_length, 1)
            position_ids_q = op.Unsqueeze(position_ids_1d, [0])
            position_ids_k = op.Unsqueeze(position_ids_1d, [0])

            query_BHSDh_rope = msft_op.RotaryEmbedding(
                query_BHSDh_normalized,
                position_ids_q,
                cos,
                sin,
            )
            key_BHkvSDh_rope = msft_op.RotaryEmbedding(
                key_BHkvSDh_normalized,
                position_ids_k,
                cos,
                sin,
            )

            if with_past:
                key_seq_BHkvSkvDh = op.Concat(past_key, key_BHkvSDh_rope, axis=-2)
                value_seq_BHkvSkvDh = op.Concat(past_value, value_BHkvSDh, axis=-2)
            else:
                key_seq_BHkvSkvDh = key_BHkvSDh_rope
                value_seq_BHkvSkvDh = value_BHkvSDh

            # Now, expand from shared heads to all heads
            key_BHkv1SDh = op.Unsqueeze(key_seq_BHkvSkvDh, [2])
            key_BHkvGSDh = op.Expand(key_BHkv1SDh, shape_BHkvGSDh)
            key_BHSDh = op.Reshape(key_BHkvGSDh, shape_BHSDh)

            value_BHkv1SDh = op.Unsqueeze(value_seq_BHkvSkvDh, [2])
            value_BHkvGSDh = op.Expand(value_BHkv1SDh, shape_BHkvGSDh)
            value_BHSDh = op.Reshape(value_BHkvGSDh, shape_BHSDh)

            # Generate causal mask:
            # where every row looks like [0, 0, ..., /*diagonal=*/ 0, minval, minval, ...]
            seq_len = op.Shape(query, end=2, start=1)
            seq_len_0D = op.Squeeze(seq_len)

            past_seq_len_0D = op.Squeeze(past_seq_length)

            total_seq_len_0D = op.Add(past_seq_len_0D, seq_len_0D)
            total_seq_len = op.Reshape(total_seq_len_0D, [-1])

            # The Phi modeling code generates the following +1 as the target-length, which seems
            # unnecessary in this context. But duplicating same logic here.
            total_seq_len_plus_1_0D = op.Add(total_seq_len_0D, 1)
            total_seq_len_plus_1 = op.Reshape(total_seq_len_plus_1_0D, [-1])

            current_range = op.Range(past_seq_len_0D, total_seq_len_0D, 1)
            mask_shape = op.Concat(seq_len, total_seq_len_plus_1, axis=0)
            min_val = op.Constant(value=minval_tp)
            mask_all_min = op.Expand(min_val, mask_shape)
            total_range_as_row = op.Range(0, total_seq_len_plus_1_0D, 1)
            current_range_as_column = op.Reshape(current_range, [-1, 1])
            boolean_mask = op.Greater(total_range_as_row, current_range_as_column)
            float_0_1_mask = op.Cast(boolean_mask, to=1)
            float_0_min_mask = op.Mul(mask_all_min, float_0_1_mask)
            mask_4d = op.Unsqueeze(float_0_min_mask, [0, 1])
            shape_B111 = op.Concat(B, plus_1, plus_1, plus_1, axis=0)
            mask_B1ST_plus = op.Expand(mask_4d, shape_B111)

            # Get rid of the extra +1 added above: total_seq_len is enough, no
            # need for total_seq_len+1.
            mask_B1ST = op.Slice(mask_B1ST_plus, [0], total_seq_len, [3], [1])

            # Now, compute attention:
            key_transposed = op.Transpose(key_BHSDh, perm=[0, 1, 3, 2])
            divisor = op.Constant(value_float=scale_factor)
            scaled_query = op.Div(query_BHSDh_rope, divisor)
            scaled_key = op.Div(key_transposed, divisor)
            attn_score = op.MatMul(scaled_query, scaled_key)
            masked_attn_score = op.Add(attn_score, mask_B1ST)
            attn_weight = op.Softmax(masked_attn_score, axis=-1)
            attention_BHSDh = op.MatMul(attn_weight, value_BHSDh)

            # Reshape back to BSD format
            attention_BSHDh = op.Transpose(attention_BHSDh, perm=[0, 2, 1, 3])
            attention_BSD = op.Reshape(attention_BSHDh, shape_BSD)

            return attention_BSD, key_seq_BHkvSkvDh, value_seq_BHkvSkvDh

        return gqa

    def test_fusion(self):
        """Test that GQA fusion is successful on source model and produces an equivalent model."""
        inputs = self.inputs

        source_model = self.source_model_script().to_model_proto(
            input_types=self.input_types,
            output_types=self.output_types,
        )
        session = ort.InferenceSession(
            source_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        source_model_outputs = session.run(None, inputs)

        # Some shapes need to be present in input model for fusion to be successful.
        # (i) Shape inference doesn't handle handle ORT contrib ops.
        # (ii) TODO: investigate if Reshape(..., ["B", "S", -1, Dh]) handled precisely
        # by shape inference.
        query_BHSDh_rope_value_info = onnx.helper.make_tensor_value_info(
            "query_BHSDh_rope",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.seqlen, self.head_size],
        )
        key_BHkvSDh_rope_value_info = onnx.helper.make_tensor_value_info(
            "key_BHkvSDh_rope",
            onnx.TensorProto.FLOAT,
            ["B", self.kv_num_heads, self.seqlen, self.head_size],
        )
        query_BSHDh_value_info = onnx.helper.make_tensor_value_info(
            "query_BSHDh",
            onnx.TensorProto.FLOAT,
            ["B", self.seqlen, self.num_heads, self.head_size],
        )
        key_BHSDh_value_info = onnx.helper.make_tensor_value_info(
            "key_BHSDh",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.total_seqlen, self.head_size],
        )
        key_BSHkvDh_value_info = onnx.helper.make_tensor_value_info(
            "key_BSHkvDh",
            onnx.TensorProto.FLOAT,
            ["B", self.seqlen, self.kv_num_heads, self.head_size],
        )
        key_transposed_value_info = onnx.helper.make_tensor_value_info(
            "key_transposed",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.head_size, self.total_seqlen],
        )
        value_BHSDh_value_info = onnx.helper.make_tensor_value_info(
            "value_BHSDh",
            onnx.TensorProto.FLOAT,
            ["B", self.num_heads, self.total_seqlen, self.head_size],
        )
        source_model.graph.value_info.extend(
            [
                query_BHSDh_rope_value_info,
                key_BHkvSDh_rope_value_info,
                query_BSHDh_value_info,
                key_BHSDh_value_info,
                key_BSHkvDh_value_info,
                key_transposed_value_info,
                value_BHSDh_value_info,
            ]
        )

        source_model_ir = ir.serde.from_proto(source_model)
        inferred_model = shape_inference.infer_shapes(source_model_ir)
        onnxscript.optimizer.optimize(inferred_model)

        count = fuse_sdpa(inferred_model, debug=True)
        self.assertGreater(count, 0)

        count = fuse_gqa(inferred_model, debug=True)
        self.assertGreater(count, 0)

        fused_model = ir.serde.to_proto(inferred_model)
        session = ort.InferenceSession(
            fused_model.SerializeToString(), providers=("CPUExecutionProvider",)
        )
        outputs3 = session.run(None, inputs)

        self.assertEqual(len(outputs3), len(source_model_outputs))
        assert_allclose(outputs3, source_model_outputs)


class GQAFusionTest2(unittest.TestCase):
    @unittest.skip("Needs too much memory.")
    def test_phi4lm(self):
        test_case = phi4lm_test()
        model = test_case.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        optimize_for_ort(model, debug=True)
        gqa_nodes = [n for n in model.graph if n.op_type == "GQA"]
        self.assertEqual(len(gqa_nodes), 2, "Expected 2i GQA nodes after fusion")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

from __future__ import annotations

from typing import Any, Iterator, overload

from onnxscript._internal import builder as _builder
from onnxscript.nn._module import Module


class ModuleList(Module):
    """Holds child modules in a list, mirroring ``torch.nn.ModuleList``.

    Children are registered with string keys ``"0"``, ``"1"``, etc., so that
    hierarchical parameter names use ``.0.``, ``.1.`` separators just like
    PyTorch.

    Example::

        class MyModel(Module):
            def __init__(self):
                super().__init__("model")
                self.layers = ModuleList([Linear(4, 4) for _ in range(3)])

            def forward(self, op, x):
                for layer in self.layers:
                    x = layer(op, x)
                return x

        # Parameters will be named:
        #   model.layers.0.weight, model.layers.1.weight, model.layers.2.weight
    """

    def __init__(self, modules: list[Module] | None = None) -> None:
        super().__init__()
        if modules is not None:
            for idx, module in enumerate(modules):
                self._register_child(str(idx), module)

    def _set_name(self, name: str) -> None:
        """Set this container's name and prefix all children's names."""
        object.__setattr__(self, "_name", name)
        for key, child in self._modules.items():
            child._set_name(f"{name}.{key}")  # pylint: disable=protected-access

    def _register_child(self, key: str, module: Module) -> None:
        """Register a child module under the given string key."""
        if module._name is None:  # pylint: disable=protected-access
            # Qualify with parent name if already set (e.g. after append)
            if self._name is not None:
                module._set_name(f"{self._name}.{key}")  # pylint: disable=protected-access
            else:
                object.__setattr__(module, "_name", key)
        self._modules[key] = module
        object.__setattr__(self, key, module)

    def append(self, module: Module) -> ModuleList:
        """Append a module to the end of the list."""
        key = str(len(self._modules))
        self._register_child(key, module)
        return self

    def extend(self, modules: list[Module]) -> ModuleList:
        """Append modules from an iterable to the end of the list."""
        for module in modules:
            self.append(module)
        return self

    @overload
    def __getitem__(self, idx: int) -> Module: ...

    @overload
    def __getitem__(self, idx: slice) -> ModuleList: ...

    def __getitem__(self, idx: int | slice) -> Module | ModuleList:
        if isinstance(idx, slice):
            keys = list(self._modules.keys())[idx]
            new_list = ModuleList()
            for i, key in enumerate(keys):
                new_list._register_child(str(i), self._modules[key])
            return new_list
        if idx < 0:
            idx += len(self._modules)
        key = str(idx)
        if key not in self._modules:
            raise IndexError(f"index {idx} is out of range")
        return self._modules[key]

    def __len__(self) -> int:
        return len(self._modules)

    def __iter__(self) -> Iterator[Module]:
        return iter(self._modules.values())

    def forward(self, op: _builder.OpBuilder, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError(
            "ModuleList is not callable directly. "
            "Iterate over its children and call them individually."
        )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
"""Fuses BatchNormalization nodes into preceding nodes. Supported fusion patterns:
- BatchNormalization ∘ Conv         -> Conv
- BatchNormalization ∘ ConvTranspose -> ConvTranspose
- BatchNormalization ∘ Gemm         -> Gemm

Approach:
    Given an inbound operation output: Y = W * X + B
    And a BatchNormalization outputs: Y_BN = (gamma * (Y - μ) / std) + β, where std = sqrt(var + eps)

    The fusion updates the inbound weights as follows:
        - W_fused = W * (gamma / std)
        - B_fused = (B - μ) * (gamma / std) + β
"""

from abc import ABC
from typing import ClassVar, Mapping

import numpy as np

from onnxscript import ir
from onnxscript.rewriter._basics import MatchResult
from onnxscript.rewriter._rewrite_rule import RewriteRuleClassBase, RewriteRuleSet


def _reshape_for_broadcast(x: np.ndarray, rank: int, axis: int = 1) -> np.ndarray:
    # Build shape: 1s everywhere except -1 at the target axis
    broadcast_shape = [1 if axis != i else -1 for i in range(rank)]
    return np.reshape(x, broadcast_shape)


class _FuseBatchNormBase(RewriteRuleClassBase, ABC):
    """Interface for BatchNormalization nodes fusion."""

    def get_filters_axis(self, attributes: Mapping[str, ir.Attr]) -> int:
        """Return the axis along which BatchNorm scale should be broadcasted."""
        raise NotImplementedError()

    def _scale_weights(
        self,
        weights: np.ndarray,
        scale_factor: np.ndarray,
        attributes: Mapping[str, ir.Attr],
    ) -> np.ndarray:
        axis = self.get_filters_axis(attributes)
        return weights * _reshape_for_broadcast(scale_factor, weights.ndim, axis=axis)

    def rewrite(self, op, x: ir.Value, inbound_out: ir.Value, batchnorm_out: ir.Value):
        batchnorm_node = batchnorm_out.producer()
        # Get BatchNorm parameters
        gamma, beta, input_mean, input_var = [
            inp.const_value.numpy() for inp in batchnorm_node.inputs[1:]
        ]

        # 1e-5 is the default value for epsilon according to
        # https://onnx.ai/onnx/operators/onnx__BatchNormalization.html#attributes
        default_eps = ir.Attr("epsilon", ir.AttributeType.FLOAT, 1e-5)
        eps = batchnorm_node.attributes.get("epsilon", default_eps).as_float()

        # Compute the scale_factor to update the inbound weights and bias
        scale_factor = gamma / np.sqrt(input_var + eps)

        # Update inbound weights
        inbound_node = inbound_out.producer()
        weights = inbound_node.inputs[1].const_value.numpy()

        fused_weights = ir.tensor(
            self._scale_weights(weights, scale_factor, inbound_node.attributes)
        )

        # Update bias
        if len(inbound_node.inputs) > 2:
            original_bias = inbound_node.inputs[2].const_value.numpy()
            bias_name = inbound_node.inputs[2].name
        else:
            original_bias = np.zeros_like(input_mean)
            # Use inbound input 1 (should be weight) to derive a name for the bias
            # to avoid name collision on initializer creation when there are multiple patterns
            # sharing the same parent nodes.
            bias_name = inbound_node.inputs[1].name + "_bias"
        fused_bias = ir.tensor((original_bias - input_mean) * scale_factor + beta)

        return op.op(
            self.op_type,
            inputs=[
                x,
                op.initializer(fused_weights, name=inbound_node.inputs[1].name),
                op.initializer(fused_bias, name=bias_name),
            ],
            attributes=inbound_node.attributes,
        )

    def check(self, context, x, inbound_out: ir.Value, batchnorm_out: ir.Value) -> MatchResult:
        del context  # Unused
        check_result = MatchResult()

        inbound_node = inbound_out.producer()
        batchnorm_node = batchnorm_out.producer()

        # Check that inbound weights + (inbound bias) + batchnorm params are initializers
        # and that they are not graph inputs
        initializers = [inbound_node.inputs[1], *batchnorm_node.inputs[1:]]
        if len(inbound_node.inputs) > 2:
            initializers.append(inbound_node.inputs[2])

        for initializer in initializers:
            if not initializer.is_initializer() or initializer.const_value is None:
                return check_result.fail(f"{initializer.name} is not a constant initializer.")
            if initializer.is_graph_input():
                return check_result.fail(f"{initializer.name} is a graph input.")

        # Check that the inbound node's weight and bias initializers are not shared
        # with other nodes outside this matched pattern. When the fusion creates new
        # initializers with the same name as the original shared weights, it overwrites
        # the original initializer in the graph, leaving other nodes that reference the
        # original value with an invalid (unregistered) input.
        matched_nodes = {inbound_node, batchnorm_node}
        inbound_initializers = [inbound_node.inputs[1]]
        if len(inbound_node.inputs) > 2:
            inbound_initializers.append(inbound_node.inputs[2])
        for init_value in inbound_initializers:
            for user, _ in init_value.uses():
                if user not in matched_nodes:
                    return check_result.fail(
                        f"Initializer '{init_value.name}' is used by another node "
                        f"'{user.name}' outside the matched pattern."
                    )

        return check_result


class FuseBatchNormIntoConv(_FuseBatchNormBase):
    """Replaces ``BatchNormalization(Conv(x))`` with ``Conv(x)``."""

    op_type: ClassVar = "Conv"

    def get_filters_axis(self, attributes: Mapping[str, ir.Attr]) -> int:
        return 0

    def pattern(self, op, x):
        return op.BatchNormalization(
            op.Conv(x, _allow_other_inputs=True, _outputs=["inbound_out"]),
            _allow_other_inputs=True,
            _outputs=["batchnorm_out"],
        )


class FuseBatchNormIntoConvTranspose(_FuseBatchNormBase):
    """Replaces ``BatchNormalization(ConvTranspose(x))`` with ``ConvTranspose(x)``."""

    op_type: ClassVar = "ConvTranspose"

    def _scale_weights(
        self,
        weights: np.ndarray,
        scale_factor: np.ndarray,
        attributes: Mapping[str, ir.Attr],
    ) -> np.ndarray:
        # ConvTranspose weight: (in_channels, out_channels/group, *kernel)
        # Reshape weights: [in_channels, out_channels/group, *kernel] → [group, in_channels/group, out_channels/group, *kernel]
        in_channels = weights.shape[0]
        out_channels_per_group = weights.shape[1]
        kernel_shape = weights.shape[2:]
        group = attributes.get("group", ir.AttrInt64("group", 1)).as_int()
        w = weights.reshape(group, in_channels // group, out_channels_per_group, *kernel_shape)

        # Per group scale_factor (out_channels,) -> (group, out_channels/group) -> (group, 1, out_channels/group, 1, ..., 1)
        s = scale_factor.reshape((group, out_channels_per_group) + (1,) * len(kernel_shape))
        # insert in_channels/group axis -> (group, 1, out_channels/group, *ones)
        s = s[:, None, ...]

        return (w * s).reshape(weights.shape)

    def pattern(self, op, x):
        return op.BatchNormalization(
            op.ConvTranspose(x, _allow_other_inputs=True, _outputs=["inbound_out"]),
            _allow_other_inputs=True,
            _outputs=["batchnorm_out"],
        )

    def check(self, context, x, inbound_out, batchnorm_out):
        check_result = super().check(context, x, inbound_out, batchnorm_out)
        if not check_result:
            return check_result

        inbound_node = inbound_out.producer()

        in_channels = inbound_node.inputs[1].const_value.numpy().shape[0]
        group = inbound_node.attributes.get("group", ir.AttrInt64("group", 1)).as_int()

        # Check that in_channels is divisible by group as ONNX checker allows it
        # But this is invalid case
        if in_channels % group != 0:
            return check_result.fail(
                f"ConvTranspose in_channels ({in_channels}) is not divisible by group ({group})."
            )

        return check_result


class FuseBatchNormIntoGemm(_FuseBatchNormBase):
    """Replaces ``BatchNormalization(Gemm(x))`` with ``Gemm(x)``."""

    op_type: ClassVar = "Gemm"

    def get_filters_axis(self, attributes: Mapping[str, ir.Attr]) -> int:
        return (
            0 if attributes.get("transB") is not None and attributes["transB"].as_int() else 1
        )

    def pattern(self, op, x):
        return op.BatchNormalization(
            op.Gemm(x, _allow_other_inputs=True, _outputs=["inbound_out"]),
            _allow_other_inputs=True,
            _outputs=["batchnorm_out"],
        )


fuse_batchnorm_into_conv_rule = FuseBatchNormIntoConv().rule()
fuse_batchnorm_into_conv_transpose_rule = FuseBatchNormIntoConvTranspose().rule()
fuse_batchnorm_into_gemm_rule = FuseBatchNormIntoGemm().rule()


rules = RewriteRuleSet(
    [
        fuse_batchnorm_into_conv_rule,
        fuse_batchnorm_into_conv_transpose_rule,
        fuse_batchnorm_into_gemm_rule,
    ]
)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
from __future__ import annotations

import unittest

from parameterized import parameterized

import onnxscript.optimizer
from onnxscript.rewriter.models import _rotary_embedding_models, _smollm_1
from onnxscript.rewriter.ort_fusions._test_utils import assert_allclose, ort_run
from onnxscript.rewriter.ort_fusions.cos_sin_cache import fuse_cos_sin_cache
from onnxscript.rewriter.ort_fusions.rotary_embedding import (
    fuse_partial_rotary_embedding,
    fuse_rotary_embedding,
)


class TestCosSinCacheTransform(unittest.TestCase):
    @parameterized.expand(
        [
            (
                "smollm_test_1",
                _smollm_1.smollm_test_1,
            ),
            (
                "test_case_1",
                _rotary_embedding_models.test_case_1,
            ),
            (
                "test_case_2",
                _rotary_embedding_models.test_case_2,
            ),
            (
                "partial_rotary_test_case",
                _rotary_embedding_models.partial_rotary_test_case,
            ),
        ]
    )
    def test_cos_sin_fusion(self, name, test_data_constructor):
        test = test_data_constructor()
        model = test.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        inputs = test.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)
        count = fuse_rotary_embedding(model)
        self.assertGreater(count, 0)
        count = fuse_cos_sin_cache(model, debug=True)
        self.assertGreater(count, 0)
        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)

    def test_partial_rotary_fusion(self):
        test = _rotary_embedding_models.partial_rotary_test_case()
        model = test.get_onnx_model()
        onnxscript.optimizer.optimize(model)
        inputs = test.get_ort_inputs()
        original_outputs = ort_run("original", model, inputs)
        count = fuse_rotary_embedding(model)
        self.assertGreater(count, 0)
        count = fuse_cos_sin_cache(model)
        self.assertGreater(count, 0)
        count = fuse_partial_rotary_embedding(model)
        self.assertGreater(count, 0)
        new_outputs = ort_run("optimized", model, inputs)
        assert_allclose(new_outputs, original_outputs)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.

# Generated from Phi2LM 1 Layer ONNX model produced by the new (Dynamo) exporter
# ruff: noqa: F821

import numpy
import onnx_ir as ir

from onnxscript import script
from onnxscript.onnx_opset import opset18
from onnxscript.onnx_types import BOOL, FLOAT, INT64

value_infos = {
    "model_embed_tokens_weight": FLOAT[51200, 2560],
    "model_layers_0_self_attn_q_proj_weight": FLOAT[2560, 2560],
    "model_layers_0_self_attn_q_proj_bias": FLOAT[2560],
    "model_layers_0_self_attn_k_proj_weight": FLOAT[2560, 2560],
    "model_layers_0_self_attn_k_proj_bias": FLOAT[2560],
    "model_layers_0_self_attn_v_proj_weight": FLOAT[2560, 2560],
    "model_layers_0_self_attn_v_proj_bias": FLOAT[2560],
    "model_layers_0_self_attn_dense_weight": FLOAT[2560, 2560],
    "model_layers_0_self_attn_dense_bias": FLOAT[2560],
    "model_layers_0_mlp_fc1_weight": FLOAT[10240, 2560],
    "model_layers_0_mlp_fc1_bias": FLOAT[10240],
    "model_layers_0_mlp_fc2_weight": FLOAT[2560, 10240],
    "model_layers_0_mlp_fc2_bias": FLOAT[2560],
    "model_layers_0_input_layernorm_weight": FLOAT[2560],
    "model_layers_0_input_layernorm_bias": FLOAT[2560],
    "model_final_layernorm_weight": FLOAT[2560],
    "model_final_layernorm_bias": FLOAT[2560],
    "lm_head_weight": FLOAT[51200, 2560],
    "lm_head_bias": FLOAT[51200],
    "expand_2": FLOAT[1, 16, 1],
    "val_1": INT64[1],
    "sym_size_int_44": INT64,
    "val_4": INT64[1],
    "val_5": INT64[1],
    "sym_size_int_50": INT64,
    "embedding": FLOAT["s34", "s16", 2560],
    "add_4": INT64,
    "val_6": FLOAT,
    "val_7": INT64,
    "arange": INT64["s16"],
    "val_8": INT64[1],
    "unsqueeze": INT64[1, "s16"],
    "val_10": FLOAT,
    "val_13": INT64[1],
    "val_14": INT64[1],
    "val_15": INT64[2],
    "full": FLOAT["s16", "s16 + s62"],
    "diagonal": INT64,
    "triu": FLOAT["s16", "s16 + s62"],
    "val_18": INT64,
    "val_19": INT64,
    "arange_1": INT64["s16 + s62"],
    "val_21": INT64[2],
    "view": INT64["s16", 1],
    "gt": BOOL["s16", "s16 + s62"],
    "convert_element_type_default": FLOAT["s16", "s16 + s62"],
    "mul_16": FLOAT["s16", "s16 + s62"],
    "val_22": INT64[1],
    "val_421": INT64[2],
    "unsqueeze_4": FLOAT[1, 1, "s16", "s16 + s62"],
    "val_23": INT64,
    "val_31": INT64,
    "val_49": INT64[1],
    "val_50": INT64[4],
    "val_52": INT64[4],
    "expand_1": FLOAT["s34", 1, "s16", "s16 + s62"],
    "val_61": INT64,
    "val_72": INT64[1],
    "val_74": INT64[1],
    "val_75": INT64[1],
    "val_78": INT64[1],
    "val_79": INT64[1],
    "slice_8": FLOAT["s34", 1, "s16", "s16 + s62"],
    "val_422": INT64[2],
    "unsqueeze_6": INT64["s34", 1, 1, "s16 + s62"],
    "convert_element_type_default_1": FLOAT["s34", 1, 1, "s16 + s62"],
    "add_89": FLOAT["s34", 1, "s16", "s16 + s62"],
    "scalar_tensor_default": FLOAT,
    "eq_64": BOOL["s34", 1, "s16", "s16 + s62"],
    "val_119": INT64[1],
    "val_121": INT64[1],
    "val_122": INT64[1],
    "val_125": INT64[1],
    "val_126": INT64[1],
    "slice_14": FLOAT["s34", 1, "s16", "s16 + s62"],
    "val_127": FLOAT,
    "masked_fill": FLOAT["s34", 1, "s16", "s16 + s62"],
    "val_179": INT64[4],
    "val_180": INT64,
    "val_181": INT64[None],
    "val_186": INT64[None, 1],
    "val_187": FLOAT["s16", 1, "s34", "s16 + s62"],
    "val_188": FLOAT["s16", 1, "s34", "s16 + s62"],
    "val_189": FLOAT["s16", 1, "s34", "s16 + s62"],
    "val_191": INT64[4],
    "val_192": INT64,
    "val_193": INT64[None],
    "val_198": INT64[None, 1],
    "val_199": FLOAT[1, "s34", "s16", "s16 + s62"],
    "val_200": FLOAT[1, "s34", "s16", "s16 + s62"],
    "val_201": FLOAT[1, "s34", "s16", "s16 + s62"],
    "slice_scatter_1": FLOAT["s34", 1, "s16", "s16 + s62"],
    "val_203": INT64[4],
    "val_204": INT64,
    "val_205": INT64[None],
    "val_210": INT64[None, 1],
    "slice_scatter_2": FLOAT["s34", 1, "s16", "s16 + s62"],
    "unsqueeze_9": INT64[1, 1, "s16"],
    "_to_copy": FLOAT[1, 1, "s16"],
    "matmul": FLOAT[1, 16, "s16"],
    "transpose": FLOAT[1, "s16", 16],
    "cat": FLOAT[1, "s16", 32],
    "cos": FLOAT[1, "s16", 32],
    "sin": FLOAT[1, "s16", 32],
    "layer_norm": FLOAT["s34", "s16", 2560],
    "val_246": FLOAT[2560, 2560],
    "val_247": FLOAT["s34", "s16", 2560],
    "linear": FLOAT["s34", "s16", 2560],
    "val_252": INT64[1],
    "val_253": INT64[4],
    "view_1": FLOAT["s34", "s16", 32, 80],
    "transpose_1": FLOAT["s34", 32, "s16", 80],
    "val_255": FLOAT[2560, 2560],
    "val_256": FLOAT["s34", "s16", 2560],
    "linear_1": FLOAT["s34", "s16", 2560],
    "val_261": INT64[4],
    "view_2": FLOAT["s34", "s16", 32, 80],
    "transpose_2": FLOAT["s34", 32, "s16", 80],
    "val_263": FLOAT[2560, 2560],
    "val_264": FLOAT["s34", "s16", 2560],
    "linear_2": FLOAT["s34", "s16", 2560],
    "val_269": INT64[4],
    "view_3": FLOAT["s34", "s16", 32, 80],
    "transpose_3": FLOAT["s34", 32, "s16", 80],
    "val_273": INT64[1],
    "val_277": INT64[1],
    "val_280": INT64[1],
    "val_281": INT64[1],
    "slice_26": FLOAT["s34", 32, "s16", 32],
    "val_284": INT64[1],
    "val_287": INT64[1],
    "val_290": INT64[1],
    "val_291": INT64[1],
    "slice_27": FLOAT["s34", 32, "s16", 48],
    "val_294": INT64[1],
    "val_297": INT64[1],
    "val_300": INT64[1],
    "val_301": INT64[1],
    "slice_28": FLOAT["s34", 32, "s16", 32],
    "val_304": INT64[1],
    "val_307": INT64[1],
    "val_310": INT64[1],
    "val_311": INT64[1],
    "slice_29": FLOAT["s34", 32, "s16", 48],
    "unsqueeze_10": FLOAT[1, 1, "s16", 32],
    "unsqueeze_11": FLOAT[1, 1, "s16", 32],
    "mul_213": FLOAT["s34", 32, "s16", 32],
    "val_314": INT64[1],
    "val_318": INT64[1],
    "val_321": INT64[1],
    "val_322": INT64[1],
    "slice_30": FLOAT["s34", 32, "s16", 16],
    "val_325": INT64[1],
    "val_328": INT64[1],
    "val_331": INT64[1],
    "val_332": INT64[1],
    "slice_31": FLOAT["s34", 32, "s16", 16],
    "neg": FLOAT["s34", 32, "s16", 16],
    "cat_1": FLOAT["s34", 32, "s16", 32],
    "mul_230": FLOAT["s34", 32, "s16", 32],
    "add_290": FLOAT["s34", 32, "s16", 32],
    "mul_238": FLOAT["s34", 32, "s16", 32],
    "val_335": INT64[1],
    "val_338": INT64[1],
    "val_341": INT64[1],
    "val_342": INT64[1],
    "slice_32": FLOAT["s34", 32, "s16", 16],
    "val_345": INT64[1],
    "val_348": INT64[1],
    "val_351": INT64[1],
    "val_352": INT64[1],
    "slice_33": FLOAT["s34", 32, "s16", 16],
    "neg_1": FLOAT["s34", 32, "s16", 16],
    "cat_2": FLOAT["s34", 32, "s16", 32],
    "mul_255": FLOAT["s34", 32, "s16", 32],
    "add_326": FLOAT["s34", 32, "s16", 32],
    "cat_3": FLOAT["s34", 32, "s16", 80],
    "cat_4": FLOAT["s34", 32, "s16", 80],
    "transpose_4": FLOAT["s34", 32, 80, "s16 + s62"],
    "matmul_1": FLOAT["s34", 32, "s16", "s16 + s62"],
    "val_353": FLOAT,
    "mul_287": FLOAT["s34", 32, "s16", "s16 + s62"],
    "val_372": INT64[1],
    "val_374": INT64[1],
    "val_375": INT64[1],
    "val_378": INT64[1],
    "val_379": INT64[1],
    "slice_41": FLOAT["s34", 1, "s16", "s16 + s62"],
    "add_387": FLOAT["s34", 32, "s16", "s16 + s62"],
    "val_380": FLOAT["s34", 32, "s16", "s16 + s62"],
    "matmul_2": FLOAT["s34", 32, "s16", 80],
    "transpose_5": FLOAT["s34", "s16", 32, 80],
    "val_385": INT64[3],
    "view_4": FLOAT["s34", "s16", 2560],
    "val_387": FLOAT[2560, 2560],
    "val_388": FLOAT["s34", "s16", 2560],
    "linear_3": FLOAT["s34", "s16", 2560],
    "val_389": FLOAT[2560, 10240],
    "val_390": FLOAT["s34", "s16", 10240],
    "linear_4": FLOAT["s34", "s16", 10240],
    "val_391": FLOAT,
    "mul_351": FLOAT["s34", "s16", 10240],
    "val_392": FLOAT,
    "pow_1": FLOAT["s34", "s16", 10240],
    "val_393": FLOAT,
    "mul_358": FLOAT["s34", "s16", 10240],
    "add_446": FLOAT["s34", "s16", 10240],
    "val_394": FLOAT,
    "mul_365": FLOAT["s34", "s16", 10240],
    "tanh": FLOAT["s34", "s16", 10240],
    "add_459": FLOAT["s34", "s16", 10240],
    "mul_375": FLOAT["s34", "s16", 10240],
    "val_395": FLOAT[10240, 2560],
    "val_396": FLOAT["s34", "s16", 2560],
    "linear_5": FLOAT["s34", "s16", 2560],
    "add_476": FLOAT["s34", "s16", 2560],
    "add_481": FLOAT["s34", "s16", 2560],
    "layer_norm_1": FLOAT["s34", "s16", 2560],
    "val_419": FLOAT[2560, 51200],
    "val_420": FLOAT["s34", "s16", 51200],
}


def make_model(
    model_embed_tokens_weight,
    model_layers_0_self_attn_q_proj_weight,
    model_layers_0_self_attn_q_proj_bias,
    model_layers_0_self_attn_k_proj_weight,
    model_layers_0_self_attn_k_proj_bias,
    model_layers_0_self_attn_v_proj_weight,
    model_layers_0_self_attn_v_proj_bias,
    model_layers_0_self_attn_dense_weight,
    model_layers_0_self_attn_dense_bias,
    model_layers_0_mlp_fc1_weight,
    model_layers_0_mlp_fc1_bias,
    model_layers_0_mlp_fc2_weight,
    model_layers_0_mlp_fc2_bias,
    model_layers_0_input_layernorm_weight,
    model_layers_0_input_layernorm_bias,
    model_final_layernorm_weight,
    model_final_layernorm_bias,
    lm_head_weight,
    lm_head_bias,
    expand_2,
):
    @script()
    def main_graph(
        input_ids: INT64["s34", "s16"],
        attention_mask: INT64["s34", "s16 + s62"],
        past_key_values_key_cache_0: FLOAT["s34", 32, "s62", 80],
        past_key_values_value_cache_0: FLOAT["s34", 32, "s62", 80],
    ) -> (
        FLOAT["s34", "s16", 51200],
        FLOAT["s34", 32, "s16 + s62", 80],
        FLOAT["s34", 32, "s16 + s62", 80],
    ):
        val_1 = opset18.Shape(input_ids, end=2, start=1)
        sym_size_int_44 = opset18.Squeeze(val_1)
        val_4 = opset18.Shape(past_key_values_value_cache_0, end=1, start=0)
        val_5 = opset18.Shape(past_key_values_value_cache_0, end=3, start=2)
        sym_size_int_50 = opset18.Squeeze(val_5)
        embedding = opset18.Gather(model_embed_tokens_weight, input_ids, axis=0)
        add_4 = opset18.Add(sym_size_int_50, sym_size_int_44)
        arange = opset18.Range(sym_size_int_50, add_4, 1)
        unsqueeze = opset18.Unsqueeze(arange, [0])
        val_14 = opset18.Reshape(add_4, [-1], allowzero=0)
        val_15 = opset18.Concat(val_1, val_14, axis=0)
        full = opset18.Expand(-3.4028235e38, val_15)
        diagonal = opset18.Constant(value_int=1)
        triu = opset18.Trilu(full, diagonal, upper=1)
        arange_1 = opset18.Range(0, add_4, 1)
        view = opset18.Reshape(arange, [-1, 1], allowzero=1)
        gt = opset18.Greater(arange_1, view)
        convert_element_type_default = opset18.Cast(gt, to=1)
        mul_16 = opset18.Mul(triu, convert_element_type_default)
        unsqueeze_4 = opset18.Unsqueeze(mul_16, [0, 1])
        val_50 = opset18.Concat(val_4, [1], [-1], [-1], axis=0)
        val_52 = opset18.Abs(val_50)
        expand_1 = opset18.Expand(unsqueeze_4, val_52)
        val_72 = opset18.Constant(value_ints=[0])
        val_74 = opset18.Constant(value_ints=[-1])
        val_75 = opset18.Reshape(add_4, val_74, allowzero=0)
        val_79 = opset18.Constant(value_ints=[1])
        slice_8 = opset18.Slice(expand_1, val_72, val_75, [3], val_79)
        unsqueeze_6 = opset18.Unsqueeze(attention_mask, [1, 2])
        convert_element_type_default_1 = opset18.Cast(unsqueeze_6, to=1)
        add_89 = opset18.Add(slice_8, convert_element_type_default_1)
        eq_64 = opset18.Equal(add_89, 0.0)
        val_119 = opset18.Constant(value_ints=[0])
        val_121 = opset18.Constant(value_ints=[-1])
        val_122 = opset18.Reshape(add_4, val_121, allowzero=0)
        val_126 = opset18.Constant(value_ints=[1])
        slice_14 = opset18.Slice(expand_1, val_119, val_122, [3], val_126)
        masked_fill = opset18.Where(eq_64, -3.4028235e38, slice_14)
        val_179 = opset18.Shape(expand_1, start=0)
        val_180 = opset18.Gather(val_179, 2, axis=0)
        val_181 = opset18.Range(0, val_180, 1)
        val_186 = opset18.Unsqueeze(val_181, [-1])
        val_187 = opset18.Transpose(masked_fill, perm=[2, 1, 0, 3])
        val_188 = opset18.Transpose(expand_1, perm=[2, 1, 0, 3])
        val_189 = opset18.ScatterND(val_188, val_186, val_187, reduction="none")
        val_191 = opset18.Shape(expand_1, start=0)
        val_192 = opset18.Gather(val_191, 1, axis=0)
        val_193 = opset18.Range(0, val_192, 1)
        val_198 = opset18.Unsqueeze(val_193, [-1])
        val_199 = opset18.Transpose(val_189, perm=[1, 2, 0, 3])
        val_200 = opset18.Transpose(expand_1, perm=[1, 0, 2, 3])
        val_201 = opset18.ScatterND(val_200, val_198, val_199, reduction="none")
        slice_scatter_1 = opset18.Transpose(val_201, perm=[1, 0, 2, 3])
        val_203 = opset18.Shape(expand_1, start=0)
        val_204 = opset18.Gather(val_203, 0, axis=0)
        val_205 = opset18.Range(0, val_204, 1)
        val_210 = opset18.Unsqueeze(val_205, [-1])
        slice_scatter_2 = opset18.ScatterND(
            expand_1, val_210, slice_scatter_1, reduction="none"
        )
        unsqueeze_9 = opset18.Unsqueeze(unsqueeze, [1])
        _to_copy = opset18.Cast(unsqueeze_9, to=1)
        matmul = opset18.MatMul(expand_2, _to_copy)
        transpose = opset18.Transpose(matmul, perm=[0, 2, 1])
        cat = opset18.Concat(transpose, transpose, axis=-1)
        cos = opset18.Cos(cat)
        sin = opset18.Sin(cat)
        layer_norm = opset18.LayerNormalization(
            embedding,
            model_layers_0_input_layernorm_weight,
            model_layers_0_input_layernorm_bias,
            stash_type=1,
            epsilon=9.999999747378752e-06,
            axis=-1,
        )
        val_246 = opset18.Transpose(model_layers_0_self_attn_q_proj_weight, perm=[1, 0])
        val_247 = opset18.MatMul(layer_norm, val_246)
        linear = opset18.Add(val_247, model_layers_0_self_attn_q_proj_bias)
        val_253 = opset18.Concat(val_4, val_1, [-1], [80], axis=0)
        view_1 = opset18.Reshape(linear, val_253, allowzero=1)
        transpose_1 = opset18.Transpose(view_1, perm=[0, 2, 1, 3])
        val_255 = opset18.Transpose(model_layers_0_self_attn_k_proj_weight, perm=[1, 0])
        val_256 = opset18.MatMul(layer_norm, val_255)
        linear_1 = opset18.Add(val_256, model_layers_0_self_attn_k_proj_bias)
        val_261 = opset18.Concat(val_4, val_1, [-1], [80], axis=0)
        view_2 = opset18.Reshape(linear_1, val_261, allowzero=1)
        transpose_2 = opset18.Transpose(view_2, perm=[0, 2, 1, 3])
        val_263 = opset18.Transpose(model_layers_0_self_attn_v_proj_weight, perm=[1, 0])
        val_264 = opset18.MatMul(layer_norm, val_263)
        linear_2 = opset18.Add(val_264, model_layers_0_self_attn_v_proj_bias)
        val_269 = opset18.Concat(val_4, val_1, [-1], [80], axis=0)
        view_3 = opset18.Reshape(linear_2, val_269, allowzero=1)
        transpose_3 = opset18.Transpose(view_3, perm=[0, 2, 1, 3])
        val_281 = opset18.Constant(value_ints=[1])
        slice_26 = opset18.Slice(transpose_1, [0], [32], [3], val_281)
        val_291 = opset18.Constant(value_ints=[1])
        slice_27 = opset18.Slice(transpose_1, [32], [9223372036854775807], [3], val_291)
        val_301 = opset18.Constant(value_ints=[1])
        slice_28 = opset18.Slice(transpose_2, [0], [32], [3], val_301)
        val_311 = opset18.Constant(value_ints=[1])
        slice_29 = opset18.Slice(transpose_2, [32], [9223372036854775807], [3], val_311)
        unsqueeze_10 = opset18.Unsqueeze(cos, [1])
        unsqueeze_11 = opset18.Unsqueeze(sin, [1])
        mul_213 = opset18.Mul(slice_26, unsqueeze_10)
        val_322 = opset18.Constant(value_ints=[1])
        slice_30 = opset18.Slice(slice_26, [0], [16], [3], val_322)
        val_332 = opset18.Constant(value_ints=[1])
        slice_31 = opset18.Slice(slice_26, [16], [9223372036854775807], [3], val_332)
        neg = opset18.Neg(slice_31)
        cat_1 = opset18.Concat(neg, slice_30, axis=-1)
        mul_230 = opset18.Mul(cat_1, unsqueeze_11)
        add_290 = opset18.Add(mul_213, mul_230)
        mul_238 = opset18.Mul(slice_28, unsqueeze_10)
        val_342 = opset18.Constant(value_ints=[1])
        slice_32 = opset18.Slice(slice_28, [0], [16], [3], val_342)
        val_352 = opset18.Constant(value_ints=[1])
        slice_33 = opset18.Slice(slice_28, [16], [9223372036854775807], [3], val_352)
        neg_1 = opset18.Neg(slice_33)
        cat_2 = opset18.Concat(neg_1, slice_32, axis=-1)
        mul_255 = opset18.Mul(cat_2, unsqueeze_11)
        add_326 = opset18.Add(mul_238, mul_255)
        cat_3 = opset18.Concat(add_290, slice_27, axis=-1)
        cat_4 = opset18.Concat(add_326, slice_29, axis=-1)
        cat_5 = opset18.Concat(past_key_values_key_cache_0, cat_4, axis=-2)
        cat_6 = opset18.Concat(past_key_values_value_cache_0, transpose_3, axis=-2)
        transpose_4 = opset18.Transpose(cat_5, perm=[0, 1, 3, 2])
        matmul_1 = opset18.MatMul(cat_3, transpose_4)
        mul_287 = opset18.Mul(matmul_1, 0.1118034)
        val_372 = opset18.Constant(value_ints=[0])
        val_374 = opset18.Constant(value_ints=[-1])
        val_375 = opset18.Reshape(add_4, val_374, allowzero=0)
        val_379 = opset18.Constant(value_ints=[1])
        slice_41 = opset18.Slice(slice_scatter_2, val_372, val_375, [3], val_379)
        add_387 = opset18.Add(mul_287, slice_41)
        val_380 = opset18.Softmax(add_387, axis=-1)
        matmul_2 = opset18.MatMul(val_380, cat_6)
        transpose_5 = opset18.Transpose(matmul_2, perm=[0, 2, 1, 3])
        val_385 = opset18.Concat(val_4, val_1, [-1], axis=0)
        view_4 = opset18.Reshape(transpose_5, val_385, allowzero=1)
        val_387 = opset18.Transpose(model_layers_0_self_attn_dense_weight, perm=[1, 0])
        val_388 = opset18.MatMul(view_4, val_387)
        linear_3 = opset18.Add(val_388, model_layers_0_self_attn_dense_bias)
        val_389 = opset18.Transpose(model_layers_0_mlp_fc1_weight, perm=[1, 0])
        val_390 = opset18.MatMul(layer_norm, val_389)
        linear_4 = opset18.Add(val_390, model_layers_0_mlp_fc1_bias)
        mul_351 = opset18.Mul(linear_4, 0.5)
        pow_1 = opset18.Pow(linear_4, 3.0)
        mul_358 = opset18.Mul(pow_1, 0.044715)
        add_446 = opset18.Add(linear_4, mul_358)
        mul_365 = opset18.Mul(add_446, 0.7978846)
        tanh = opset18.Tanh(mul_365)
        add_459 = opset18.Add(tanh, 1.0)
        mul_375 = opset18.Mul(mul_351, add_459)
        val_395 = opset18.Transpose(model_layers_0_mlp_fc2_weight, perm=[1, 0])
        val_396 = opset18.MatMul(mul_375, val_395)
        linear_5 = opset18.Add(val_396, model_layers_0_mlp_fc2_bias)
        add_476 = opset18.Add(linear_3, linear_5)
        add_481 = opset18.Add(add_476, embedding)
        layer_norm_1 = opset18.LayerNormalization(
            add_481,
            model_final_layernorm_weight,
            model_final_layernorm_bias,
            stash_type=1,
            epsilon=9.999999747378752e-06,
            axis=-1,
        )
        val_419 = opset18.Transpose(lm_head_weight, perm=[1, 0])
        val_420 = opset18.MatMul(layer_norm_1, val_419)
        linear_6 = opset18.Add(val_420, lm_head_bias)
        return linear_6, cat_5, cat_6

    model = main_graph.to_model_proto(value_infos=value_infos)
    return model


def make_model_with_random_weights():
    model_embed_tokens_weight = numpy.random.rand(51200, 2560).astype(numpy.float32)
    model_layers_0_self_attn_q_proj_weight = numpy.random.rand(2560, 2560).astype(
        numpy.float32
    )
    model_layers_0_self_attn_q_proj_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_self_attn_k_proj_weight = numpy.random.rand(2560, 2560).astype(
        numpy.float32
    )
    model_layers_0_self_attn_k_proj_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_self_attn_v_proj_weight = numpy.random.rand(2560, 2560).astype(
        numpy.float32
    )
    model_layers_0_self_attn_v_proj_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_self_attn_dense_weight = numpy.random.rand(2560, 2560).astype(numpy.float32)
    model_layers_0_self_attn_dense_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_mlp_fc1_weight = numpy.random.rand(10240, 2560).astype(numpy.float32)
    model_layers_0_mlp_fc1_bias = numpy.random.rand(10240).astype(numpy.float32)
    model_layers_0_mlp_fc2_weight = numpy.random.rand(2560, 10240).astype(numpy.float32)
    model_layers_0_mlp_fc2_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_input_layernorm_weight = numpy.random.rand(2560).astype(numpy.float32)
    model_layers_0_input_layernorm_bias = numpy.random.rand(2560).astype(numpy.float32)
    model_final_layernorm_weight = numpy.random.rand(2560).astype(numpy.float32)
    model_final_layernorm_bias = numpy.random.rand(2560).astype(numpy.float32)
    lm_head_weight = numpy.random.rand(51200, 2560).astype(numpy.float32)
    lm_head_bias = numpy.random.rand(51200).astype(numpy.float32)
    expand_2 = numpy.random.rand(1, 16, 1).astype(numpy.float32)
    model = make_model(
        model_embed_tokens_weight,
        model_layers_0_self_attn_q_proj_weight,
        model_layers_0_self_attn_q_proj_bias,
        model_layers_0_self_attn_k_proj_weight,
        model_layers_0_self_attn_k_proj_bias,
        model_layers_0_self_attn_v_proj_weight,
        model_layers_0_self_attn_v_proj_bias,
        model_layers_0_self_attn_dense_weight,
        model_layers_0_self_attn_dense_bias,
        model_layers_0_mlp_fc1_weight,
        model_layers_0_mlp_fc1_bias,
        model_layers_0_mlp_fc2_weight,
        model_layers_0_mlp_fc2_bias,
        model_layers_0_input_layernorm_weight,
        model_layers_0_input_layernorm_bias,
        model_final_layernorm_weight,
        model_final_layernorm_bias,
        lm_head_weight,
        lm_head_bias,
        expand_2,
    )
    return model


class _Phi2LMTest:
    def get_onnx_model(self):
        if not hasattr(self, "_onnx_model"):
            model_proto = make_model_with_random_weights()
            model = ir.serde.deserialize_model(model_proto)
            self._onnx_model = model
        return self._onnx_model


def phi2lm_test():
    return _Phi2LMTest()
